"""Module defines the `Ambient` class. This class encapsulates the state of the environment, see the class docs for details."""

from __future__ import annotations
from typing import Any, TYPE_CHECKING
from abc import ABC, abstractmethod
import ray

from ..utils import int64_uuid, _Future
from ..agent import _Agent
from ..event import Event, Action, ActiveObservation, ErrorActiveObservation
from ..pubsub import Subscribe, Unsubscribe

if TYPE_CHECKING:
    from ..agent import Agent


class Ambient(ABC):
    """Base class for `Ambient` implementations.

    The `Ambient` is contains the state of the environment, that is, it contains all variables (or data) that are accessible and manipulable by the agents present in the given environment. The ambient also contains references to all of the agents that are present in the environment.

    All ambients must implement two key methods: `__select__` and `__update__`.

    `__select__` is a read-only operation which takes an `Action` and will return an `Observation`. The `Action` typically defines what data the agent is interested in observing, and the `Observation` will contain this data. These actions will originate exclusively from the agents `Sensor`s and are executed during the `__sense__` step in the agents cycle.

    `__update__` is a write operation which also takes an `Action` and may return an `Observation`. The `Action` defines how the environments state (the `Ambient`) will be mutated. These actions will originate exclusively from the agents `Actuator`s and are executed during the `__execute__` step in the agents cycle.

    Another important method `__subscribe__` is used to handle actions that are specifically related to `star_ray`s pub-sub mechansim. Not all ambients need implement this.

    Agents can be added or removed from the environent via corresponding methods in the `Ambient`.
    """

    def __init__(self, agents: list[Agent], *args, **kwargs):
        """Constructor.

        Args:
            agents (list[Agent]): a list of agents that will initially be added to this `Ambient`.
            args (list[Any]): optional additional arguments.
            kwargs (dict[str, Any]): optional additional arguments.
        """
        super().__init__(*args, **kwargs)
        self._id = int64_uuid()
        agents = [_Agent.new(agent) for agent in agents]
        self._agents = {agent.get_id(): agent for agent in agents}
        self._is_alive = False

    def add_agent(self, agent: Agent) -> _Agent:
        """Adds a new agent to this ambient.

        Args:
            agent (Any): the agent to add

        Raises:
            ValueError: if the agent already exists.

        Returns:
            _Agent: A handle (and wrapper) to the added agent.
        """
        agent = _Agent.new(agent)
        # TODO we need to explicitly prevent remote agents if the ambient is not remote itself
        # otherwise the agent will be modifying a copy of the ambient!
        agent_id = agent.get_id()
        if agent_id in self._agents:
            raise ValueError(
                f"Agent {agent_id} already exists in ambient {self.get_id()}."
            )
        self._agents[agent_id] = agent
        return agent

    def remove_agent(self, agent: Agent) -> None:
        """Removes an agent from this `Ambient`. This has the side effect of terminating the agent (via `Agent.__terminate__`).

        Args:
            agent (Agent): the agent to remove.

        """
        agent = _Agent.new(agent)
        agent_id = agent.get_id()
        del self._agents[agent_id]
        agent.__terminate__()

    @property
    def is_alive(self) -> bool:
        """Whether this `Ambient` is currently alive - otherwise it has shutdown (via `__terminate__`) and all agents have either terminated or are in the process of terminating. This will also return False if the `Ambient` has not yet been initialised via `__initialise__`.

        Returns:
            bool: whether this ambient is alive.
        """
        return self._is_alive

    @property
    def id(self) -> int:
        """The id of this `Ambient`.

        Returns:
            int : the unique identifier of this `Ambient`.
        """
        return self._id

    def get_is_alive(self) -> bool:
        """Getter for `is_alive`, see property for details.

        Returns:
            bool: whether this ambient is alive.
        """
        return self._is_alive

    def get_id(self) -> int:
        """Getter for `id`, see property for details.

        Returns:
            int: the unique identifier of this `Ambient`
        """
        return self._id

    def get_agents(self) -> list[_Agent]:
        """Getter for all agents that are currently present in this `Ambient`. This is read-only and modifications to the returned list will have no effect. If you want to add or remove an agent use the corresponding methods.

        Returns:
            list[_Agent]: list of agents.
        """
        return list(self._agents.values())

    def get_agent_count(self) -> int:
        """Get the number of agents currently in this ambient."""
        return len(self._agents)

    async def __terminate__(self) -> None:
        """Terminate this `Ambient`. After this call `is_alive` will return False. This call will wait for all agents to be terminated via their `__terminate__` method."""
        state = _Ambient.new(self)
        self._is_alive = False
        agents = list(self.get_agents())
        self._agents.clear()
        # TODO if an agent takes too long, then just cancel it?
        await _Future.gather([agent.__terminate__(state) for agent in agents])

    async def __initialise__(self) -> None:
        """Initialise this `Ambient`. After this call `is_alive` will return True. This call will wait for all agents to initialise via their `__initialise__` method."""
        self._is_alive = True
        state = _Ambient.new(self)
        # TODO if an agent takes too long, then just cancel it?
        agents = list(self.get_agents())
        await _Future.gather([agent.__initialise__(state) for agent in agents])

    @abstractmethod
    def __select__(self, action: Action) -> ActiveObservation | ErrorActiveObservation:
        """A read-only operation which takes an `Action` and will return an `Observation`. The `Action` typically defines what data the agent is interested in observing, and the `Observation` will contain this data. These actions will originate exclusively from the agents `Sensor`s and are executed during the `__sense__` step in the agents cycle. This method may be called manually (especially when dealing with actions/observations in a subclass of `Ambient`, `Environment` or `Sensor`.

        Args:
            action (Action): the sense action

        Returns:
            ActiveObservation | ErrorActiveObservation: the resulting observation
        """
        pass

    @abstractmethod
    def __update__(
        self, action: Action
    ) -> ActiveObservation | ErrorActiveObservation | None:
        """A write operation which also takes an `Action` and may return an `Observation`. The `Action` defines how the environments state (this `Ambient`) will be mutated. These actions will originate exclusively from the agents `Actuator`s and are executed during the `__execute__` step in the agents cycle. This method may be called manually (especially when dealing with actions/observations in a subclass of `Ambient`, `Environment` or `Actuator`.

        Args:
            action (Action): the action

        Returns:
            ActiveObservation | ErrorActiveObservation | None: the resulting observation (or None)
        """
        pass

    def __subscribe__(
        self, action: Subscribe | Unsubscribe
    ) -> ActiveObservation | ErrorActiveObservation:
        """Handles subcriptions to events in this ambient if the `Ambient` supports pub-sub. By default this does nothing. If you are considering implementing pub-sub functionality, see the `star_ray.pubsub` sub-package for details.

        Typically, this method will be called from `__select__` as `Sensor`s are often interested subscribing to receive certain events are observations. It is also possible (but unusual) to call this from `__update__`

        Args:
            action (Subscribe | Unsubscribe): the subscription action

        Returns:
            ActiveObservation | ErrorActiveObservation: an observation typically indicating whether the subscription was successful
        """
        pass


class _Ambient(ABC):
    @staticmethod
    def new(ambient: Ambient | ray.actor.ActorHandle) -> _Ambient:
        if isinstance(ambient, ray.actor.ActorHandle):
            return _AmbientRemote(ambient)
        elif isinstance(ambient, Ambient):
            return _AmbientLocal(ambient)
        else:
            raise TypeError(type(ambient))

    @property
    @abstractmethod
    def is_alive(self):
        pass

    @abstractmethod
    async def __initialise__(self):
        pass

    @abstractmethod
    async def __terminate__(self):
        pass

    @abstractmethod
    def __subscribe__(self, actions: list[Event]) -> list[Any]:
        pass

    @abstractmethod
    def __update__(self, actions: list[Event]) -> list[Any]:
        pass

    @abstractmethod
    def __select__(self, actions: list[Event]) -> list[Any]:
        pass

    @abstractmethod
    def get_agents(self) -> list[_Agent]:
        pass

    @abstractmethod
    def get_agent_count(self) -> int:
        pass


class _AmbientRemote(_Ambient):
    def __init__(self, ambient: ray.actor.ActorHandle):
        super().__init__()
        self._inner = ambient

    @property
    def is_alive(self):
        return self._inner.get_is_alive.remote()

    async def __initialise__(self):
        return await self._inner.__initialise__.remote(self)

    async def __terminate__(self):
        return await self._inner.__terminate__.remote(self)

    def __subscribe__(self, actions: list[Subscribe | Unsubscribe]) -> list[Any]:
        return [self._inner.__subscribe__.remote(query) for query in actions]

    def __update__(self, actions: list[Event]) -> list[Any]:
        return [self._inner.__update__.remote(query) for query in actions]

    def __select__(self, actions: list[Event]) -> list[Any]:
        return [self._inner.__select__.remote(query) for query in actions]

    def get_agents(self) -> list[_Agent]:
        return ray.get(self._inner.get_agents.remote())

    def get_agent_count(self) -> int:
        return ray.get(self._inner.get_agent_count.remote())


class _AmbientLocal(_Ambient):
    def __init__(self, ambient: Ambient):
        super().__init__()
        self._inner = ambient

    @property
    def is_alive(self):
        return self._inner.get_is_alive()

    async def __initialise__(self):
        return await self._inner.__initialise__()

    async def __terminate__(self):
        return await self._inner.__terminate__()

    def __subscribe__(self, actions: list[Subscribe | Unsubscribe]) -> list[Any]:
        return [self._inner.__subscribe__(query) for query in actions]

    def __update__(self, actions: list[Event]) -> list[Any]:
        return [self._inner.__update__(query) for query in actions]

    def __select__(self, actions: list[Event]) -> list[Any]:
        return [self._inner.__select__(query) for query in actions]

    def get_agents(self) -> list[_Agent]:
        return self._inner.get_agents()

    def get_agent_count(self) -> int:
        return self._inner.get_agent_count()



"""This module defines events base classes, such as `Event`, `Observation`, `Action` and various user input events."""

from .event import Event
from .user_event import (
    KeyEvent,
    MouseButtonEvent,
    MouseMotionEvent,
    JoyStickEvent,
    WindowFocusEvent,
    WindowMoveEvent,
    WindowResizeEvent,
    WindowCloseEvent,
    WindowOpenEvent,
    ScreenSizeEvent,
)
from .action_event import Action
from .observation_event import (
    Observation,
    ActiveObservation,
    ErrorActiveObservation,
    ErrorObservation,
    wrap_observation,
)

__all__ = (
    "Event",
    "Action",
    "Observation",
    "ActiveObservation",
    "ErrorActiveObservation",
    "ErrorObservation",
    # user input events
    "KeyEvent",
    "JoyStickEvent",
    "MouseButtonEvent",
    "MouseMotionEvent",
    "WindowFocusEvent",
    "WindowMoveEvent",
    "WindowResizeEvent",
    "WindowCloseEvent",
    "WindowOpenEvent",
    "ScreenSizeEvent",
    # other
    "wrap_observation",
)



"""Module defining the `KeyEvent` class, it is used to represent user keyboard input."""

from typing import ClassVar
from pydantic import field_validator
from ..event import Event


class KeyEvent(Event):
    """A class representing a keyboard event.

    Attributes:
        id ([int]): A unique identifier for the event.
        timestamp ([float]): The timestamp (in seconds since UNIX epoch) when the event instance is created.
        source ([int]): A unique identifier for the source of this event.
        key ([str]): The string representation of the key.
        keycode ([int]): The key code of the keyboard event.
        status ([int]): The status of the key event (UP = 0, DOWN = 1, HOLD = 2).
    """

    key: str
    keycode: int
    status: int

    UP: ClassVar[int] = 0
    DOWN: ClassVar[int] = 1
    HOLD: ClassVar[int] = 2

    @field_validator("status", mode="before")
    def _validate_status(cls, value: str | int):
        if isinstance(value, str):
            if "release" in value or value == "up":
                return cls.UP
            elif "press" in value or value == "down":
                return cls.DOWN
            elif "hold" in value:
                return cls.HOLD
            else:
                raise ValueError(
                    f"Failed to convert status string: {value}, valid values include: ['released', 'pressed', 'clicked', 'up', 'down']"
                )
        elif isinstance(value, int) and value in [
            KeyEvent.UP,
            KeyEvent.DOWN,
            KeyEvent.HOLD,
        ]:
            return value
        else:
            raise ValueError(f"Invalid {type(KeyEvent)} status code: {value}")



import json
import copy
import re

from pathlib import Path
from typing import (
    Any,
    Literal,
    Optional,
)
from collections.abc import Mapping, MutableMapping, Sequence, Callable
from jinja2 import (
    BytecodeCache,
    Environment,
    BaseLoader,
    PrefixLoader,
    PackageLoader,
    FileSystemLoader,
    Template,
    TemplateNotFound,
    Undefined,
    StrictUndefined,
)
from jinja2.runtime import Context
from jinja2.defaults import (
    BLOCK_END_STRING,
    BLOCK_START_STRING,
    COMMENT_END_STRING,
    COMMENT_START_STRING,
    KEEP_TRAILING_NEWLINE,
    LSTRIP_BLOCKS,
    TRIM_BLOCKS,
    VARIABLE_END_STRING,
    VARIABLE_START_STRING,
    LINE_STATEMENT_PREFIX,
    LINE_COMMENT_PREFIX,
    NEWLINE_SEQUENCE,
)
from jinja2.ext import Extension
from starlette.templating import Jinja2Templates


from deepmerge import always_merger
from cerberus import Validator as _Validator

from ._logging import _LOGGER as LOGGER

__all__ = (
    "Validator",
    "TemplateLoader",
    "ValidatedTemplates",
    "ValidatedTemplate",
    "ValidatedEnvironment",
)


class Validator(_Validator):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("normalize", True)
        super().__init__(*args, **kwargs)
        self._hex_color_pattern = re.compile(r"^#?[0-9a-fA-F]{6}$")

    def _validate_type_color(self, value):
        return re.match(self._hex_color_pattern, value) is not None

    def _validate_type_any(self, _):
        return True


class ValidatedTemplates(Jinja2Templates):
    def __init__(self):
        loader = TemplateLoader()
        ALLOW_REQUEST = {"request": {"type": "any", "default": ""}}
        env = ValidatedEnvironment(loader=loader, schema_globals=ALLOW_REQUEST)
        super().__init__(env=env)

    def add_namespace(
        self,
        namespace: str,
        path: str | list[str],
        package_name: str = None,
        follow_links: bool = False,
    ):
        self.loader.add_namespace(
            namespace, path, package_name=package_name, follow_links=follow_links
        )

    def TemplateResponse(self, *args, **kwargs):
        return super().TemplateResponse(*args, **kwargs)

    @property
    def loader(self) -> "TemplateLoader":
        loader = self.env.loader
        assert isinstance(loader, TemplateLoader)
        return loader


class ValidatedTemplate(Template):
    def new_context(
        self,
        vars: dict[str, Any] | None = None,
        shared: bool = False,
        locals: Mapping[str, Any] | None = None,
    ) -> Context:
        assert isinstance(self.environment, ValidatedEnvironment)
        self.environment.validate_context(self.name, vars)
        return super().new_context(vars, shared, locals)


class ValidatedEnvironment(Environment):
    EXT_SCHEMA = ".schema.json"
    EXT_CONTEXT = ".json"

    allow_unknown = False
    validator_class = Validator
    template_class = ValidatedTemplate

    def __init__(
        self,
        block_start_string: str = BLOCK_START_STRING,
        block_end_string: str = BLOCK_END_STRING,
        variable_start_string: str = VARIABLE_START_STRING,
        variable_end_string: str = VARIABLE_END_STRING,
        comment_start_string: str = COMMENT_START_STRING,
        comment_end_string: str = COMMENT_END_STRING,
        line_statement_prefix: str | None = LINE_STATEMENT_PREFIX,
        line_comment_prefix: str | None = LINE_COMMENT_PREFIX,
        trim_blocks: bool = TRIM_BLOCKS,
        lstrip_blocks: bool = LSTRIP_BLOCKS,
        newline_sequence: Literal["\\n", "\\r\\n", "\\r"] = NEWLINE_SEQUENCE,
        keep_trailing_newline: bool = KEEP_TRAILING_NEWLINE,
        extensions: Sequence[str | type["Extension"]] = (),
        optimized: bool = True,
        undefined: type[
            Undefined
        ] = StrictUndefined,  # defaults to StrictUndefined now for better validation!
        finalize: Callable[..., Any] | None = None,
        autoescape: bool | Callable[[str | None], bool] = False,
        loader: Optional["BaseLoader"] = None,
        cache_size: int = 400,
        auto_reload: bool = True,
        bytecode_cache: Optional["BytecodeCache"] = None,
        enable_async: bool = False,
        schema_globals: Mapping[str, Any] = None,
    ):
        super().__init__(
            block_start_string,
            block_end_string,
            variable_start_string,
            variable_end_string,
            comment_start_string,
            comment_end_string,
            line_statement_prefix,
            line_comment_prefix,
            trim_blocks,
            lstrip_blocks,
            newline_sequence,
            keep_trailing_newline,
            extensions,
            optimized,
            undefined,
            finalize,
            autoescape,
            loader,
            cache_size,
            auto_reload,
            bytecode_cache,
            enable_async,
        )
        self._schema_globals = schema_globals if schema_globals else dict()
        self._validator_cache = dict()

    def get_template(
        self,
        name: str | Template,
        parent: str | None = None,
        globals: MutableMapping[str, Any] | None = None,
    ) -> Template:
        if name not in self._validator_cache:
            self._validator_cache[name] = None

        filename, _ = ValidatedEnvironment.split_name_suffix(name)
        # check for configuration files with the same name
        context_name: str = filename.with_suffix(
            ValidatedEnvironment.EXT_CONTEXT
        ).as_posix()
        schema_name: str = filename.with_suffix(
            ValidatedEnvironment.EXT_SCHEMA
        ).as_posix()
        try:
            schema = json.loads(super().get_template(schema_name, None, {}).render())
            schema = always_merger.merge(copy.deepcopy(self._schema_globals), schema)
            validator = ValidatedEnvironment.get_validator(schema)
            self._validator_cache[name] = validator
        except TemplateNotFound:
            schema = None
        try:
            context = json.loads(super().get_template(context_name, None, {}).render())
        except TemplateNotFound:
            context = None
        if schema:
            context = ValidatedEnvironment._validate_context_with_schema(
                validator, context=context, normalize=True
            )
        elif context:
            LOGGER.warning(
                f"Context file: {context_name} was provided without a validation schema.",
            )
            context = ValidatedEnvironment._validate_context_without_schema(context)
        else:
            context = dict()  # no configuration files were available
        globals = always_merger.merge(globals, context)
        # this will be used when the template is rendered (via `get_context`) to validate supplied context variables.

        return super().get_template(name, parent, globals)

    def validate_context(self, name: str, context: Mapping[str, Any]):
        validator = self._validator_cache.get(name, None)
        if validator:
            ValidatedEnvironment._validate_context_with_schema(
                validator, context=context, normalize=False
            )
        else:
            ValidatedEnvironment._validate_context_without_schema(context)

    @staticmethod
    def _validate_context_without_schema(context: Mapping[str, Any]):
        errors = ValidatedEnvironment._validate_context_keys(context)
        if errors:
            error_str = "\n    - ".join(errors)
            raise ValueError(f"Invalid context, see errors:\n    - {error_str}")
        return context

    @staticmethod
    def _validate_context_with_schema(
        validator: Any,
        context: MutableMapping[str, Any] = None,
        normalize: bool = True,
    ):
        if context is None:
            context = dict()
        if normalize:
            context = validator.normalized(context)  # set default values etc.
            if context is None:  # there were errors in normalization
                errors = ValidatedEnvironment._format_validator_errors(validator.errors)
                error_str = "\n    - ".join(errors)
                raise ValueError(
                    f"Context is not valid under the provided schema. See issues below:\n    - {error_str}"
                )
        if validator.validate(context):
            return context
        else:
            errors = ValidatedEnvironment._format_validator_errors(validator.errors)
            error_str = "\n    - ".join(errors)
            raise ValueError(
                f"Context is not valid under the provided schema. See issues below:\n    - {error_str}"
            )

    @staticmethod
    def _format_validator_errors(errors):
        result = []
        for k, errs in errors.items():
            for v in errs:
                if isinstance(v, dict):
                    result.extend(ValidatedEnvironment._format_validator_errors(v))
                elif isinstance(v, str):
                    result.append(f"Key: `{k}` {v}")
                else:
                    raise ValueError(
                        f"Internal error: unknown validation error type: {type(v)}"
                    )
        return result

    @staticmethod
    def get_validator(schema: MutableMapping[str, Any]) -> Validator:
        errors = ValidatedEnvironment._validate_schema(schema)
        if errors:
            error_str = "\n    - ".join(errors)
            raise ValueError(f"Invalid schema, see errors:\n    - {error_str}")
        return ValidatedEnvironment.validator_class(
            schema, normalize=True, allow_unknown=ValidatedEnvironment.allow_unknown
        )

    @staticmethod
    def load_and_validate_context(schema_path: str, context_path: str = None):
        schema_path = Path(schema_path).expanduser().resolve().absolute().as_posix()
        with open(schema_path, encoding="utf-8") as f:
            validator = ValidatedEnvironment.get_validator(json.load(f))
        context = None
        if context_path:
            with open(context_path, encoding="utf-8") as f:
                context = json.load(f)
        return ValidatedEnvironment._validate_context_with_schema(
            validator,
            context=context,
            normalize=True,
        )

    @staticmethod
    def split_name_suffix(file: str):
        filename = Path(file).name
        parent = Path(file).parent
        name, *suffix = filename.split(".")
        return parent / name, suffix

    @staticmethod
    def _validate_context_keys(context: MutableMapping[str, Any]):
        errors = []
        for key, _ in context.items():
            # TODO check that the key follows jinja variable syntax
            if "-" in key:
                errors.append(f"Invalid character '-' found in context key: '{key}'.")
        return errors

    @staticmethod
    def _validate_schema(schema: MutableMapping[str, Any]):
        errors = []
        errors.extend(ValidatedEnvironment._validate_context_keys(schema))
        for key, rules in schema.items():
            if isinstance(rules, dict):
                if rules.get("type", None) == "dict":
                    # default must be specified in the nested definition
                    errors.extend(ValidatedEnvironment._validate_context_keys(schema))
                    sub_schema = rules.get("schema", None)
                    if sub_schema:
                        errors.extend(ValidatedEnvironment._validate_schema(sub_schema))
                    continue  # default isnt specified at the top level here...
                if "default" not in rules:
                    errors.append(f"No default value specified for field: '{key}'")
        return errors


class TemplateLoader(BaseLoader):
    def __init__(self):
        super().__init__()
        self._prefix_loader = PrefixLoader({})

    def add_namespace(
        self,
        namespace: str,
        path: str | list[str],
        *args,
        package_name: str = None,
        follow_links: bool = False,
        **kwargs,
    ):
        assert isinstance(namespace, str)
        assert isinstance(path, list | str)
        if isinstance(path, list):
            assert all(isinstance(p, str) for p in path)
        if package_name:
            self._prefix_loader.mapping[namespace] = PackageLoader(
                package_name, package_path=path
            )
        else:
            self._prefix_loader.mapping[namespace] = FileSystemLoader(
                path, followlinks=follow_links
            )

    def get_loader(self, namespace: str) -> tuple[BaseLoader, str]:
        return self._prefix_loader.get_loader(namespace + "/")[0]

    def get_namespaces(self) -> list[str]:
        return list(self._prefix_loader.mapping.keys())

    def list_templates_in_namespace(self, namespace) -> list[str]:
        return self._prefix_loader.get_loader(namespace + "/")[0].list_templates()

    def list_templates(self) -> list[str]:
        return self._prefix_loader.list_templates()

    def get_source(self, environment, template):
        return self._prefix_loader.get_source(environment, template)



from collections import defaultdict

from ._pubsub import Subscriber
from ._pubsub import Publisher


class TopicPublisher(Publisher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._subscribers: dict[str, set[Subscriber]] = defaultdict(set)

    def subscribe(self, topic: str, subscriber: Subscriber) -> None:
        self._subscribers[topic].add(subscriber)

    def unsubscribe(self, topic: str, subscriber: Subscriber) -> None:
        self._subscribers[topic].discard(subscriber)



import types


__all__ = ("SliceType", "EllipsisType")


class SliceType:
    """This type is used in place of the built-in `slice` type to be compatible with pydantic validation.

    TODO: create a pull request in pydantic to support the `slice` type by default (as well as the `Ellipsis` type)
    """

    def __init__(self, start: int, stop: int, step: int = None):
        self.start = start
        self.stop = stop
        self.step = step

    def get_value(self) -> slice:
        return slice(self.start, self.stop, self.step)


class EllipsisType:
    def get_value(self) -> types.EllipsisType:
        return ...



"""Module contains events related to application window changes.

Typically there will only be one such window as part of a simulation and so no details of the window (e.g. title) are provided. If this is required, extend the classes present here.
"""

from ..event import Event


class WindowMoveEvent(Event):
    """Event indicating that an application window was moved."""

    position: tuple[int, int] | tuple[float, float]


class WindowResizeEvent(Event):
    """Event indicating that an application window was resized."""

    size: tuple[int, int] | tuple[float, float]


class WindowCloseEvent(Event):
    """Event indicating that an application window was closed."""

    pass


class WindowOpenEvent(Event):
    """Event indicating that an application window was opened."""

    pass


class WindowFocusEvent(Event):
    """Event representing a change in an application window focus."""

    has_focus: bool


class ScreenSizeEvent(Event):
    """Event indicating that the screen/monitor has a given size."""

    size: tuple[float, float] | tuple[int, int]



"""Module defines the decorator `@attempt`.

The `@attempt` decorator should be used on methods that are meant to take actions in `Actuator` and `Sensor` components. It is also used in the automatic routing of actions to relevant components as implemented in the `AgentRouted` class. See these classes and the `attempt` documentation for details.
"""

from functools import wraps
from ...utils import TypeRouter

__all__ = ("attempt",)


def attempt(*event_types: list[type]):
    """A decorator that defines an `attempt` method within a `Component`. `attempt` methods should be called in the agent's `__cycle__` method to schedule an action for execution. This is true of both actuators and sensors.

    Args:
        event_types (List[Type]): A list of types that the decorated `Component` method will receive if event routing is enabled. If this is not specified then an attempt will be made to discover the types via the type hint of the first argument of the decorated function.

    Sensors: `Sensor` actions will be attempted at the start of the next cycle (before the next `__cycle__` is called).
    Actuators: `Actuator` actions will be attempted at the end of the agents current cycle (after `__cycle__` has finished).

    For an action the properly scheduled for execution it must be returned by the decorated method (see example below).

    Example:
    ```
    class MyActuator(Actuator):
        @attempt
        def move(self, direction):
            return MoveAction(direction)


    class MyAgent(Agent):
        def __cycle__(self):
            # the action will be executed after __cycle__ is finished.
            self.actuators[0].move("EAST")
    ```


    """

    def _attempt(func, route_types=None):
        if route_types is None:
            route_types = TypeRouter.resolve_first_argument_types(
                func, allow_no_arguments=True
            )
        else:
            route_types = TypeRouter.resolve_route_types(route_types)

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            action = func(self, *args, **kwargs)
            if isinstance(action, list | tuple):
                self._actions.extend(action)
            else:
                self._actions.append(action)
            return action

        # used to identify whether a given method has been decorated with this decorator.
        wrapper.is_attempt = True
        # this is used to automatically route events to a given component.
        if route_types:
            route_types = TypeRouter.validate_route_types(func, route_types)
            wrapper.route_types = route_types
        return wrapper

    if len(event_types) == 0:
        return _attempt
    elif len(event_types) == 1 and isinstance(event_types[0], list | tuple):

        def _wrap(func, route_types=event_types[0]):
            return _attempt(func, route_types=route_types)

        return _wrap
    elif len(event_types) == 1:
        # event_types is actually the function we are wrapping... used the decorator as @attempt without arguments
        return _attempt(event_types[0], route_types=None)
    else:
        raise ValueError(
            "Decorator `@attempt` takes a single positional argument `event_types`."
        )



"""Module contains base classes for observations, see classes: `Observation`, `ActiveObservation`, `ErrorObservation`, `ErrorActiveObservation` documentation for details."""

import traceback
from functools import wraps
from typing import Any
from pydantic import field_validator, Field
from .event import Event


class Observation(Event):
    """Base class for an observation. Contains `values` (any type) which holds the observed data."""

    values: Any = Field(default_factory=lambda: None)


class ActiveObservation(Observation):
    """Base class for an observation that was the result of an action. Contains `action_id` which is the unique `id` of the action that generated this observation."""

    action_id: int

    @field_validator("action_id", mode="before")
    def _validate_query_id(cls, value):
        if isinstance(value, Event):
            return value.id
        elif isinstance(value, int):
            return value
        else:
            raise ValueError(f"Invalid action_id {value}")

    @staticmethod
    def new(action: Event, values: Any) -> "ActiveObservation":
        """Factory method.

        Args:
            action (Event): action that lead to this observation.
            values (Any): values that are part of this observation.

        Returns:
            ActiveObservation: the observation.
        """
        return ActiveObservation(action_id=action, values=values)


class ErrorObservation(Observation):
    """Base class for an observation that contains an error.

    This typically means an exception was raised during some computation that is related to the observation. It holds relevant exception information, the exception can be re-raised using: `raise observation.exception()`.

    The factory method `from_exception` is the easiest way to create this kind of observation in a try-catch block.
    ```
    try:
        ...
    catch Exception as e:
        return ErrorObservation.from_exception(e)
    ```
    """

    exception_type: str
    traceback_message: str

    def __str__(self):  # noqa: D105
        return f"ErrorResponse(\nsource={self.source},\n{self.traceback_message}\n)"

    def exception(self):
        """Creates an exception that may be re-raised, the exception contains information about the original error.

        Returns:
            _ObservationError: the exception to re-raise (or process)
        """
        return _ObservationError(f"\n{self.exception_type}:\n{self.traceback_message}")

    def from_exception(exception: Exception) -> "ErrorObservation":
        """Factory method that will build an instance from an `Exception`.

        Args:
            exception (Exception): the exception
        Returns:
            ErrorActiveObservation: the error observation
        """
        exception_type = exception.__class__.__name__
        traceback_message = "".join(
            traceback.format_exception(
                type(exception), value=exception, tb=exception.__traceback__
            )
        )
        return ErrorObservation(
            exception_type=exception_type,
            traceback_message=traceback_message,
        )


class ErrorActiveObservation(ErrorObservation, ActiveObservation):
    """An observation that contains an error and was the result of an action."""

    def from_exception(action: Event, exception: Exception) -> "ErrorActiveObservation":
        """Factory for `ErrorActiveObservation` that will build an instance from an `Exception`.

        Args:
            action (Event): the action that led to the exception
            exception (Exception): the exception

        Returns:
            ErrorActiveObservation: the error observation
        """
        exception_type = exception.__class__.__name__
        traceback_message = "".join(
            traceback.format_exception(
                type(exception), value=exception, tb=exception.__traceback__
            )
        )
        return ErrorActiveObservation(
            action_id=action,
            exception_type=exception_type,
            traceback_message=traceback_message,
        )


class _ObservationError(Exception):
    """Wrapper exception class that will contain information about an exception that occured during observation computation."""


def wrap_observation(fun):
    """Decorator that will wrap the return value in an ActiveObservation or ErrorActiveObservation if there was an exception. Assumes that `action` is the first argument of the given function.

    Args:
        fun: function to decorate.
    """

    # TODO check that the first argument is an action!
    @wraps(fun)
    def _wrap(
        self, action: Event, *args, **kwargs
    ) -> ActiveObservation | ErrorActiveObservation:
        try:
            result = fun(self, action, *args, **kwargs)
            if not isinstance(result, ActiveObservation):
                return ActiveObservation.new(action, values=result)
            else:
                return result
        except Exception as e:
            return ErrorActiveObservation.from_exception(action, e)

    return _wrap



"""Module defining the class `Component`, see class documentation for details."""

from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from ...event import Observation, Action
from .._wrapper_observations import _Observations
from ...utils import int64_uuid

if TYPE_CHECKING:
    from ..agent import Agent
    from ...environment import State

__all__ = ("Component",)


class Component(ABC):
    """Base class for components that may be attached to an agent. These include: `Sensor` and `Actuator`, which are used by the agent for sensing the environment and performing actions to change its state."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        super().__init__(*args, **kwargs)
        self._id: int = int64_uuid()
        # this will be set by the agent when this component is added to it.
        # generally it should not be accessed directly unless you know what you are doing!
        self._agent = None
        # actions to attempt in the current cycle to produce observations
        self._actions: list[Action] = []  # TODO allow async access here?
        # observations that result from taking action
        self._observations: _Observations = _Observations.empty()

    def on_add(self, agent: Agent) -> None:
        """Callback for when this `Component` is added to an `Agent`.

        Args:
            agent (Agent): agent that this `Component` was added to.
        """
        self._agent = agent

    def on_remove(self, agent: Agent) -> None:
        """Callback for when this `Component` is removed from an `Agent`.

        Args:
            agent (Agent): agent that this `Component` was removed from to.
        """
        self._agent = None

    @property
    def id(self):
        """Unique identifier for this [`Component`].

        Returns:
            [str]: unique identifier
        """
        return self._id

    def iter_observations(self):
        """Iterate over and consumes the observations that are currently stored in this `Component`.

        Yields:
            [Observation]: the most recent observation.
        """
        yield from filter(None, self._observations)

    def iter_actions(self):
        """Iterates (and consumes?TODO) the actions that are currently stored in this `Component`.

        Yields:
            [Action]: the most recent action.
        """
        yield from filter(None, self._actions)

    async def aiter_observations(self):
        """Not implemented."""
        raise NotImplementedError("TODO")

    async def aiter_actions(self):
        """Not implemented."""
        raise NotImplementedError("TODO")

    def __transduce__(self, events: list[Observation]) -> list[Observation]:
        """Convert a list of events to another list of events. This is typically used to implement a transformation of "raw" observations that this component may receive. There is no restiction on how this transformation is done, some events may even be removed or added.

        This should not be called manually, and is instead part of the agent's (or component's) cycle.

        Args:
            events (list[Observation]): events to transform

        Returns:
            list[Observation]: the transformed list of events
        """
        return events

    @abstractmethod
    def __query__(self, state: State) -> None:
        """Query the state of the environment (take an action).

        This should not be called manually, and is instead part of the agent's (or component's) cycle.

        Args:
            state (State): state of the environment.
        """

    def __str__(self):  # noqa: D105
        return f"{self.__class__.__name__}({self._id})"

    def __repr__(self):  # noqa: D105
        # TODO perhaps include information about the agent this component is attached to ?
        return str(self)

    @staticmethod
    def set_action_source(component: Component, actions: list[Action]):
        """Set the `source` attribute of each action in `actions` to be the `id` of this `Component`. Typically `component` is the component that is taking the actions.

        Args:
            component (Component): that is taking the actions
            actions (list[Action]): whoses sources should be set
        """
        for action in actions:
            if action.source is None:
                action.source = component.id



"""Module defines the `Agent` base class, which should be used for all agent implementations. It partially defines the agent-environment interaction API. See class documentation for details."""

from __future__ import annotations

from abc import abstractmethod, ABC
from .component import Sensor, Actuator, Component
from ..utils import int64_uuid

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..environment import State


class Agent(ABC):
    """An agent is a small program that runs in a loop, continously sensing its environment and taking actions to modify it.

    The execution of the agent is typically managed by its environment and proceedes via the following method calls:
    - `__initialise__(state)` set up the agent.
    - <LOOP BEGINS>
    - `__sense__(state)` where sensing happens.
    - `__cycle__()` where  "thinking", learning and decision making happens.
    - `__execute__(state)` where acting happens.
    - <LOOP ENDS>
    - `__terminate__(state)` tear down the agent.

    These methods should never be called by the agent itself.

    In a typical implementation, a subclass of `Agent` will implement:
    - `__initialise__`
    - `__cycle__`
    - `__terminate__`

    The `__sense__` and `__execute__` methods will handle sensing and acting via the agents `Sensor`s and `Actuator`s automatically, but they may be overriden in some advanced cases.

    ### Thinking

    Thinking happens in `__cycle__`, this is where an agent will update its beliefs and make decisions.
    A typically `__cycle__` implementation may look as follows:
    ```
    def __cycle__(self):
        # process observations and update beliefs/learn
        for sensor in self.sensors:
            for observation in sensor.iter_observations():
                pass  # update beliefs
        for actuator in self.actuators:
            for observation in sensor.iter_observations():
                pass  # update beliefs (and typically handle any errors)

        # take actions based on beliefs
        if self.beliefs["empty_ahead"]:
            self.actuators[0].move_ahead()
        else:
            self.sensors[0].empty_ahead()
        # take other actions
    ```
    Note that both sensors and actuators can take actions, and recieve obervations (as a result of actions), see below.

    ### Sensing

    An `Agent` senses via its collection of `Sensor`s, each defines "read-only" actions which are taken during the call `__sense__` before `__cycle__` is called. Observations will be gathered as a result of any actions and will typically be avaliable via the `iter_observations` method in the `Sensor`.
    Some `Sensor`s may instead define convenience methods for the `Agent` to use which extract relevant information from the most recent observations.

    ### Acting

    An `Agent` acts via its collection of `Actuator`s, each defines "write" actions which are taken during the call to `__execute__` after `__cycle__` is called. In some cases, an `Actuator` may receive feedback from the envirronment in the form of an observation. `Actuator` observations will typically contain error messages or feedback about whether a given action was successful. Whether `Actuator`s receive any feedback at all is environment dependent. Note that the observations gathered from an `Actuator` (again via `iter_observations`) are from the PREVIOUS cycle, there will always be zero `Actuator` observations on the first call to `__cycle__`.

    See also: `AgentRouted`
    """

    def __init__(
        self, sensors: list[Sensor], actuators: list[Actuator], *args, **kwargs
    ):
        """Constructor.

        Args:
            sensors (List[Sensor]): collection of sensors.
            actuators (List[Actuator]): collection of actuators.
            args (tuple[Any], optional): optional additional arguments.
            kwargs (dict[str, Any], optional): optional additional keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self._id = int64_uuid()
        self._sensors: set[Sensor] = set()
        self._actuators: set[Actuator] = set()

        self._ic: set[Component] = set()  # components to initialise
        self._tc: set[Component] = set()  # components to terminate

        self._is_initialised = False
        self._is_terminated = False

        for sensor in sensors:
            self.add_component(sensor)
        for actuator in actuators:
            self.add_component(actuator)

        if len(self.sensors) != len(sensors) or len(self.actuators) != len(actuators):
            raise ValueError(
                "Components were not added to this agent upon initialisation, did you override `add_component` and forget to call super()?"
            )

    @property
    def sensors(self) -> set[Sensor]:
        """Getter for the agent's `Sensor`s.

        Returns:
            Set[Sensor]: set of sensors.
        """
        return set(self._sensors)

    def get_sensors(self, oftype: type | None = None) -> set[Sensor]:
        """Getter for the agent's `Sensor`s.

        Arguments:
            oftype (type): type sensor(s) to get.

        Returns:
            Set[Sensor]: set of sensors.
        """
        if oftype:
            return set(filter(lambda x: isinstance(x, oftype), self._sensors))
        else:
            return set(self._sensors)

    @property
    def actuators(self) -> set[Actuator]:
        """Getter for the agent's `Actuator`s.

        Returns:
            Set[Sensor]: set of actuators.
        """
        return set(self._actuators)

    def get_actuators(self, oftype: type | None = None) -> set[Actuator]:
        """Getter for the agent's `Actuator`s.

        Arguments:
            oftype (type): type sensor(s) to get.

        Returns:
            Set[Actuator]: set of actuators.
        """
        if oftype:
            return set(filter(lambda x: isinstance(x, oftype), self._actuators))
        else:
            return set(self._actuators)

    @property
    def id(self) -> int:
        """Getter for the agent's unique id (read-only).

        Returns:
           int: the agent's id.
        """
        return self._id

    def get_id(self) -> int:
        """Getter for the agent's unique id (read-only).

        Returns:
           int: the agent's id.
        """
        return self._id

    def __initialise__(self, state: State):
        """Initialises the agent. Overriding allows set up of the agent. Initial sense actions may be triggered here to ensure the resulting observations are avaliable in the first call to `__cycle__`.

        Args:
            state (State): the state of the environment.
        """
        # TODO call query?

    def __terminate__(self, state: State):
        """Terminates the agent. Overriding allows safe clean up of any resources the agent or its `Component`s might be using. By default this will automatically remove all sensors and actuators from the agent to allow them to perform safe clean up.

        Args:
            state (State): the state of the environment.
        """
        for sensor in self.sensors:
            self.remove_component(sensor)
        for actuator in self.actuators:
            self.remove_component(actuator)
        # TODO call query?

    def __sense__(self, state: State, *args, **kwargs):
        """Take sense actions via the agents sensors. This should never be called from within the `Agent`.

        Args:
            state (State): the state of the environment.
            args (tuple[Any], optional): optional additional arguments.
            kwargs (dict[str, Any], optional): optional additional keyword arguments.
        """
        _ = [sensor.__query__(state) for sensor in self.sensors]

    @abstractmethod
    def __cycle__(self):
        """Update beliefs/learn and take actions via sensors and actuators."""

    def __execute__(self, state: State, *args, **kwargs):
        """Executes actions via the agents actuators. This should never be called from within the `Agent`.

        Args:
            state (State): the state of the environment.
            args (tuple[Any], optional): optional additional arguments.
            kwargs (dict[str, Any], optional): optional additional keyword arguments.
        """
        _ = [actuator.__query__(state) for actuator in self.actuators]

    def add_component(self, component: Component) -> Component:
        """Add a new component (sensor or actuator) to this agent.

        Args:
            component (Component): the component to add.

        Raises:
            TypeError: if the component is not of type `Sensor` or `Actuator`.

        Returns:
            Component: the component added.
        """
        if isinstance(component, Sensor):
            self._sensors.add(component)
        elif isinstance(component, Actuator):
            self._actuators.add(component)
        else:
            raise TypeError(f"Unsupported component type: {type(component)}")
        component.on_add(self)
        return component

    def remove_component(self, component: Component) -> Component:
        """Removes an existing component from this agent.

        Args:
            component (Component): component to remove.

        Raises:
            TypeError: if the component is not of type `Sensor` or `Actuator`.

        Returns:
            Component: the component that was removed.
        """
        if isinstance(component, Sensor):
            self._sensors.remove(component)
        elif isinstance(component, Actuator):
            self._actuators.remove(component)
        else:
            raise TypeError(f"Unsupported component type: {type(component)}")
        component.on_remove(self)
        return component



from weakref import ref
from abc import ABC, abstractmethod
from typing import Any
from ray.actor import ActorHandle
from copy import deepcopy

__all__ = ("Subscriber", "Publisher")


class Subscriber(ABC):
    @abstractmethod
    def __notify__(self, message: Any) -> None:
        pass


def _SubscriberWrapper(subscriber: Any):
    if isinstance(subscriber, Subscriber):
        return _SubscriberLocal(subscriber)
    elif isinstance(subscriber, ActorHandle):
        return _SubscriberRemote(subscriber)
    else:
        raise TypeError(
            f"subscriber must of type {Subscriber} or be a remote ray actor."
        )


class _SubscriberLocal(Subscriber):
    def __init__(self, subscriber: Subscriber):
        self._handle = ref(subscriber)

    def __notify__(self, message: Any) -> None:
        sub = self._handle()
        if sub:
            sub.__notify__(deepcopy(message))
        else:
            raise ValueError("Subscriber was garbage collected before unsubscribing.")

    def __eq__(self, other):
        if isinstance(other, _SubscriberRemote):
            return self._handle == other._handle
        return False

    def __hash__(self):
        return hash(self._handle)

    def __repr__(self) -> str:
        return f"Subscriber@local({repr(self._handle())})"

    def __str__(self):
        return repr(self)


class _SubscriberRemote(Subscriber):
    def __init__(self, actor_handle: ActorHandle):
        self._handle = actor_handle

    def __notify__(self, message: Any) -> None:
        self._handle.__notify__.remote(deepcopy(message))

    def __eq__(self, other):
        if isinstance(other, _SubscriberRemote):
            return self._handle == other._handle
        return False

    def __hash__(self):
        return hash(self._handle)

    def __repr__(self) -> str:
        return f"Subscriber@remote({repr(self._handle)})"

    def __str__(self):
        return str(self._handle)


class Publisher(ABC):
    @abstractmethod
    def subscribe(self, topic: Any, subscriber: Subscriber) -> None:
        pass

    @abstractmethod
    def unsubscribe(self, topic: Any, subscriber: Subscriber) -> None:
        pass

    @abstractmethod
    def publish(self, message: Any) -> None:
        pass



"""TODO."""
# import asyncio
# from abc import ABC, abstractmethod

# from .ambient import Ambient


# class Task(ABC):

#     @abstractmethod
#     async def __call__(self, ambient, *args, **kwargs):
#         pass

#     def spawn(self):
#         return asyncio.create_task(self())

# class TaskAgent(Task):

#     def __init__(self, ambient, wait=0.0):
#         self._ambient = ambient
#         self._agents = asyncio.Queue()
#         self._wait = wait

#     async def insert(self, agent, index=0):
#         await agent.__initialise__(self._ambient)
#         await self._agents.put(index, agent)

#     async def remove(self, agent):
#         await self._agents.remove(agent) # terminate?
#         await agent.__terminate__(self._ambient)

#     async def __call__(self):
#         async for agent in self._agents:
#             asyncio.sleep(self._wait)
#             await agent.__sense__(self._ambient)
#             await agent.__cycle__()
#             await agent.__execute__(self._ambient)


# class Schedule:

#     def __init__(self):
#         self._tasks = []

#     def add_agent(self, agent, ambient, wait=1.0):
#         if wait > 0.0:
#             task = Schedule._run_wait(ambient, agent, wait=wait)
#         else:
#             task = Schedule._run_no_wait(ambient, agent)
#         self._tasks.append(task)

#     @staticmethod
#     async def _run_no_wait(ambient, agent, **kwargs):
#         await agent.__initialise__(ambient)
#         while ambient.is_alive:  # check that the agent is alive...?


#     @staticmethod
#     async def _run_wait(ambient, agent, wait=1.0, **kwargs):
#         await agent.__initialise__(ambient)
#         while ambient.is_alive:  # check that the agent is alive...?
#             await asyncio.sleep(wait)
#             await agent.__sense__(ambient)
#             await agent.__cycle__()
#             await agent.__execute__(ambient)

#     # def remove_agent(self, agent):
#     # this should kill the agent

#     # async def run(self):



"""DEPRECATED Module defining the `VisibilityEvent`]class."""

from typing import ClassVar
from pydantic import field_validator
from ..event import Event


class VisibilityEvent(Event):
    """A class representing a visibility change event.

    If this event is triggered it is an indicator that the user has either:
        1. switch window (changed focus)
        2. switched tab (if using a browser based UI)
        3. minimised the UI window

    This event DOES NOT NECESSARILY indicate that the use cannot see the interface at all, this is due to inherent limitations with detecting visibility changes (e.g. in browser based UI's). See e.g. [visibilitychange](https://developer.mozilla.org/en-US/docs/Web/API/Document/visibilitychange_event)

    Attributes:
        id ([int]): A unique identifier for the event.
        timestamp ([float]): The timestamp (in seconds since UNIX epoch) when the event instance is created.
        source ([int]): A unique identifier for the source of this event.
        status ([int]): the visibility status (VISIBLE = 0, HIDDEN = 1).
    """

    status: int

    # static fields
    VISIBLE: ClassVar[int] = 0
    HIDDEN: ClassVar[int] = 1

    @field_validator("status")
    def _validate_status(cls, value: str | int):
        if isinstance(value, str):
            if "hidden" == value:
                return cls.HIDDEN
            elif "visible" == value:
                return cls.VISIBLE
            else:
                raise ValueError(
                    f"Failed to convert status string: {value}, valid values include: ['visible', 'hidden']"
                )
        elif isinstance(value, int) and value in [
            VisibilityEvent.VISIBLE,
            VisibilityEvent.HIDDEN,
        ]:
            return value
        else:
            raise ValueError(f"Invalid {type(VisibilityEvent)} status code: {value}")



"""Module defining the `TypeRouter` class. The class is used to route objects to type-compatible functions. See the class documentation for details."""

from typing import Any
from collections.abc import Callable
from collections import defaultdict
from functools import lru_cache
import inspect
import typing
import types as _types


class TypeRouter:
    """Route objects to type-compatible functions."""

    def __init__(self, cache_size: int = 1024):
        """Constructor.

        Args:
            cache_size (int, optional): the size of type mapping cache used to determine which functions should be called for a given input type (see `__call__`). Defaults to 1024.
        """
        super().__init__()
        self._router = defaultdict(list)

        @lru_cache(maxsize=cache_size)
        def _get_funcs(t: type):
            for t in t.mro():
                funcs = self._router.get(TypeRouter.fully_qualified_name(t), None)
                if funcs:
                    return funcs
            return self._router.get(TypeRouter.fully_qualified_name(typing.Any), [])

        self._get_funcs = _get_funcs

    def add(self, func: Callable, route_types: list[type] | None = None):
        """Add a new function that will be called if an instance with a type that inherits from any type in `route_types` is provided in `__call__`.

        Note: adding a new function will invalidate the type routing cache. This class is intended to be used statically (or new functions are to be added infreqently).

        Args:
            func (Callable): function to call.
            route_types (list[type] | None, optional): instance types to route. Defaults to None, in which case the types will be resolved from the type hints of the first argument of `func`.
        """
        if route_types is None:
            assert hasattr(func, "route_types")
            route_types = func.route_types
        else:
            route_types = TypeRouter.resolve_route_types(route_types)
        if route_types is None:
            raise ValueError(f"Failed to resolve `route_types` for function: {func}.")

        for t in route_types:
            fqn = TypeRouter.fully_qualified_name(t)
            self._router[fqn].append(func)
        # cache must be recomputed if we add a new type
        self._get_funcs.cache_clear()

    def __call__(self, event: Any, *args, **kwargs) -> list[Any]:
        """Routes the given `event` to type-compatible functions. If there is not exact type match, the called function is resvolved via the instance types method resolution order (mro). The function(s) that are assocaited with the inherited type that is highest in this hierarchy (i.e. the closest parent type) are called.

        Args:
            event (Any): to route, this will be provided as the first (positional) argument to the function(s).
            *args (Any, optional): additional arguments to provide to the function(s)
            **kwargs (Any, optional): additional keyword arguments to provide to the function(s)

        Returns:
            list[Any]: results of each function call (may be empty if no type-compatible functions were found).
        """
        result = []
        for func in self._get_funcs(type(event)):
            result.append(func(event, *args, **kwargs))
        return result

    @staticmethod
    def fully_qualified_name(type: type) -> str:
        """Get the fully qualified type name of the given type.

        Note that in this class types that have the same fully qualified name as assumed to be equal.

        Args:
            type (type): type to get name of

        Returns:
            str: fully qualified name of the type
        """
        return type.__module__ + "." + type.__name__

    @staticmethod
    def get_all_decorated_methods(obj: Any) -> list[Callable]:
        """Get all of the (bound) methods in the given object that have the `route_types` attribute.

        Args:
            obj (Any): object for which to get (bound) methods.

        Returns:
            list[Callable]: list of methods that have the `route_type` attribute.
        """
        methods = inspect.getmembers(obj, predicate=inspect.ismethod)
        methods = [m[1] for m in methods]
        return list(filter(lambda x: hasattr(x, "route_types"), methods))

    @staticmethod
    def resolve_route_types(types: type | list[type]) -> list[type]:
        """Resolve types from a list of types. This handles type hints and converts them into standard python types (i.e. python primitive types or class objects). It splits the `UnionType` into its consituent.

        Args:
            types (type | list[type]): types to resolve.

        Returns:
            list[type]: resolved types.
        """
        if not isinstance(types, list | tuple):
            types = [types]
        result = []
        for type in types:
            origin = typing.get_origin(type)
            if origin is None:
                result.append(type)
            elif origin == _types.UnionType or origin == typing.Union:
                result.extend(TypeRouter.resolve_route_types(typing.get_args(type)))
            else:
                result.append(origin)
        return result

    @staticmethod
    def route(*route_types: list[type]) -> Callable:
        """A decorator that will set up a function or method for use with the `TypeRouter` class, types will be resolved statically. This is intended to be used with `get_all_decorated_methods` which will discover all (bound) instance methods. The methods can then be added to an instance of `TypeRouter`.

        Args:
            route_types (list[type]): an optional list of types that will be used as the `route_types` for the given function. If this is not specified the types will be resolved from the first argument of the function (ignoring bound arguments `self` and `cls`).

        Example 1:
        ```
        class MyClass:
            @TypeRouter.route
            def foo(self, data: str | int):
                pass  # do something
        ```

        Example 2:
        ```
        class MyClass:
            @TypeRouter.route([int | str])
            def foo(self, data):
                pass  # do something
        ```

        Example 2:
        ```
        class MyClass:
            @TypeRouter.route([int, str])
            def foo(self, data):
                pass  # do something
        ```

        Note that if `route_types` is specified, this will be prefered over any type hints.
        """

        def _route(func, route_types=None):
            if route_types is None:
                route_types = TypeRouter.resolve_first_argument_types(
                    func, allow_no_arguments=True
                )
            else:
                route_types = TypeRouter.resolve_route_types(route_types)
            # this is used to automatically route events to a given component.
            if route_types:
                route_types = TypeRouter.validate_route_types(func, route_types)
                func.route_types = route_types
            return func

        if len(route_types) == 0:
            return _route
        elif len(route_types) == 1 and isinstance(route_types[0], list | tuple):

            def _wrap(func, route_types=route_types[0]):
                return _route(func, route_types=route_types)

            return _wrap
        elif len(route_types) == 1:
            # event_types is actually the function we are wrapping... used the decorator as @attempt without arguments
            return _route(route_types[0], route_types=None)
        else:
            raise ValueError(
                "Decorator `route` takes a single positional argument `route_types`."
            )

    @staticmethod
    def resolve_first_argument_types(
        func: Callable, allow_no_arguments=False
    ) -> list[type] | None:
        """Resolves the type(s) of the first argument of the given `func` from its type hints (ignoring bound arguments `self` or `cls`).

        Args:
            func (Callable): the function whose argument should be resolved
            allow_no_arguments (bool, optional): allow no arguments in the functions, if True always return None. Defaults to False.

        Raises:
            ValueError: if the function has no arguments and `allow_no_arguments=False`
            TypeError: if no type hint was defined for the first argument.

        Returns:
            list[type]: types of the first argument
        """
        # attempt to resolve these by inspecting the func signature type hints
        parameters = list(inspect.signature(func).parameters.items())
        if len(parameters) == 0:
            if not allow_no_arguments:
                raise ValueError("Function has no arguments.")
            else:
                return None
        # TODO this is not very robust...
        if parameters[0][0] in ("self", "cls"):
            parameters = parameters[1:]
        if len(parameters) == 0:
            if not allow_no_arguments:
                raise ValueError("Function has no arguments.")
            else:
                return None
        first = parameters[0][1]
        if first.annotation == inspect._empty:
            raise TypeError(
                f"A type hint was not defined for first parameter of: {func}"
            )
        return TypeRouter.resolve_route_types([first.annotation])

    @staticmethod
    def validate_route_types(func: Callable, route_types: list[type]) -> list[type]:
        """Validate the `route_types` of a function. This is a utility method used typically used to validate the arguments provided to a TypeRouter.

        Args:
            func (Callable): the function that is being used, `route_types` may be derived from the type hints of this function, or may have been provided manually with the intention of later calling `func` with instances of these types.
            route_types (list[type]): types to validate.

        Raises:
            TypeError: if the types are not validate

        Returns:
            list[type]: the validated types (same as `route_types`)
        """
        try:
            return TypeRouter.validate_types(route_types)
        except Exception as e:
            raise TypeError(f"Invalid `route_types` for function: {func}") from e

    @staticmethod
    def validate_types(types: list[type]) -> list[type]:
        """Validate the types in `types`.

        Args:
            types (list[type]): to validate
        """
        # validate the routes provided
        if not isinstance(types, list | tuple):
            raise ValueError("types must be a list or tuple.")
        if len(types) == 0:
            raise ValueError("types must not be empty.")
        for cls in types:
            if not (isinstance(cls, type | _types.UnionType) or cls is typing.Any):
                raise ValueError(
                    f"types must contain only `type` but received: {cls}",
                )
        return list(types)



"""Module defining the `Event` class."""

import time
from pydantic import BaseModel, Field
from ..utils import int64_uuid

EVENT_TIMESTAMP_FUNC = time.time
EVENT_UUID_FUNC = int64_uuid


class Event(BaseModel):
    """An event class with a unique identifier, timestamp and source.

    Attributes:
        id ([int]): A unique identifier for the event.
        timestamp ([float]): The timestamp (in seconds since UNIX epoch) when the event instance is created.
        source ([int | None]): A unique identifier for the source of this event. By default this
    """

    id: int = Field(default_factory=EVENT_UUID_FUNC)
    timestamp: float = Field(default_factory=EVENT_TIMESTAMP_FUNC)
    source: int | None = Field(default_factory=lambda: None)

    class Config:  # noqa: D106
        validate_assignment = True



from typing import Any
from functools import partial
from ._pubsub import Publisher, Subscriber
from ..utils import _LOGGER, TypeRouter


class TypePublisher(Publisher):
    def __init__(self, cache_size=1024):
        super().__init__()
        self._router = TypeRouter(cache_size=cache_size)
        self._subscribers = dict()

    def subscribe(
        self,
        topic: type | list[type],
        subscriber: Subscriber,
    ) -> None:
        _subscriber = partial(subscriber.__notify__)
        self._router.add(_subscriber, route_types=topic)
        _LOGGER.debug(f"{self} new subscription: {subscriber} to topic: {topic}")

    def unsubscribe(self, topic: Any, subscriber: Subscriber) -> None:
        raise NotImplementedError(
            "TODO - requires `remove` implementation in TypeRouter"
        )

    def publish(self, message: Any) -> None:
        self._router(message)



from typing import Any

from pydantic import field_validator, Field
import ray

from ..event import Action
from ._pubsub import Subscriber, _SubscriberLocal, _SubscriberRemote

__all__ = ("Subscribe", "Unsubscribe")


class _SubscriptionActionBase(Action):
    topic: Any
    subscriber: Subscriber | None = Field(default_factory=lambda: None)

    @field_validator("subscriber", mode="before")
    @classmethod
    def _validate_subscriber(cls, subscriber: Subscriber | None):
        if subscriber is None:
            return None
        if ray.is_initialized():
            ctx = ray.get_runtime_context()
            if ctx.worker.mode:
                return _SubscriberRemote(ctx.current_actor)
        return _SubscriberLocal(subscriber)  # uses a weak reference

    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True


class Subscribe(_SubscriptionActionBase):
    pass


class Unsubscribe(_SubscriptionActionBase):
    pass



"""Package defining common user input events."""

from .keyevent import KeyEvent
from .mouseevent import MouseButtonEvent, MouseMotionEvent
from .joystickevent import JoyStickEvent
from .windowevent import (
    WindowFocusEvent,
    WindowMoveEvent,
    WindowResizeEvent,
    WindowCloseEvent,
    WindowOpenEvent,
    ScreenSizeEvent,
)

__all__ = (
    "KeyEvent",
    "JoyStickEvent",
    "MouseButtonEvent",
    "MouseMotionEvent",
    "WindowFocusEvent",
    "WindowMoveEvent",
    "WindowResizeEvent",
    "WindowCloseEvent",
    "WindowOpenEvent",
    "ScreenSizeEvent",
)



"""Module defines the :class:`IOSensor` which represents a senors that gets its observations from a device. These sensors do not read the environment state, instead they may read a file, grab user input, or perform some other IO READ operation."""

from __future__ import annotations
from typing import Any, TYPE_CHECKING

from ...event import Event
from .sensor import Sensor

if TYPE_CHECKING:
    from ..agent import Agent
    from ...environment import State


class IOSensor(Sensor):
    """Base class for sensors whose observations are obtained from an IO device or stream. This may be a file, user input stream, or other IO device/stream.

    The sensor wraps an IO device and assumes it defines the following methods:
    - `get_nowait() -> list[Event]`
    - `start()` -> None`
    - `stop()` -> None`

    `get_nowait()` will be used to get observations at each cycle of the sensor.

    TODO an async `get() -> list[Event]` method?
    TODO perhaps use the __enter__ __exit__ pattern instead? this is more commonly defined by IO devices (e.g. files).
    TODO implement a wrapper for files and pipes (it would be nice if we could read from file streams/os pipes)
    """

    def __init__(self, device: Any):
        """Constructor.

        Args:
            device (Any): io device that will provide observations (Event) to the sensor via `get_nowait` or `get`.
        """
        super().__init__()
        # the device must contain this method TODO a more indepth check
        assert hasattr(device, "get_nowait")
        self._device = device

    def __query__(self, _: State) -> None:  # noqa
        # we query from the device, not from the state
        observations: list[Event] = self._device.get_nowait()
        # preprocess the observations ready to be received by the agent
        self._observations.push_all(self.__transduce__(observations))

    def on_add(self, agent: Agent) -> None:  # noqa
        super().on_add(agent)
        start = getattr(self._device, "start", None)
        if start:
            start()

    def on_remove(self, agent: Agent) -> None:  # noqa
        super().on_remove(agent)
        stop = getattr(self._device, "stop", None)
        if stop:
            stop()



"""Package containing various useful utilities."""

from . import error
from ._async import _Future
from ._logging import _LOGGER
from ._uuid import int64_uuid, str_uuid4

from ._types import SliceType, EllipsisType
from ._templating import ValidatedTemplates, ValidatedEnvironment, TemplateLoader
from .type_routing import TypeRouter

__all__ = (
    "TypeRouter",
    "ValidatedTemplates",
    "ValidatedEnvironment",
    "TemplateLoader",
    "SliceType",
    "EllipsisType",
    "_Future",
    "error",
    "int64_uuid",
    "str_uuid4",
    "_LOGGER",
)



"""Module contains functions to safely evaluate strings converting them to python code (or objects), see `literal_eval` and `literal_eval_with_ops`."""

from ast import (
    Constant,
    List,
    Tuple,
    Name,
    Starred,
    Dict,
    BinOp,
    UnaryOp,
    Subscript,
    UAdd,
    USub,
    Expression,
    Mod,
    FloorDiv,
    Add,
    Sub,
    Mult,
    Div,
    Call,
    Set,
    parse,
    literal_eval,
)

__all__ = ("literal_eval_with_ops", "literal_eval")


def literal_eval_with_ops(node_or_string):
    """Safely evaluate an expression node or a string containing a Python expression. The string or node provided may only consist of the following Python literal structures: strings, bytes, numbers, tuples, lists, dicts, sets, booleans, None, and basic binary operations.

    Caution: A complex expression can overflow the C stack and cause a crash.
    """

    def _raise_malformed_node(node):
        msg = "malformed node or string"
        if lno := getattr(node, "lineno", None):
            msg += f" on line {lno}"
        raise ValueError(msg + f": {node!r}")

    def _convert(node):
        if isinstance(node, Constant):
            return node.value
        elif isinstance(node, Tuple):
            return tuple(map(_convert, node.elts))
        elif isinstance(node, List):
            # Handle list with unpacking
            result = []
            for elt in node.elts:
                if isinstance(elt, Starred):  # Unpacking operator
                    result.extend(_convert(elt.value))
                else:
                    result.append(_convert(elt))
            return result
        elif isinstance(node, Set):
            return set(map(_convert, node.elts))
        elif isinstance(node, Call):
            # Valid function calls and their handlers
            valid_calls = {
                "min": lambda args: min(_convert(arg) for arg in args),
                "max": lambda args: max(_convert(arg) for arg in args),
                "set": lambda args: set(_convert(arg) for arg in args),
            }
            if isinstance(node.func, Name) and node.func.id in valid_calls:
                # Check if the function name is valid
                if node.args != node.keywords == []:
                    # Ensure no additional arguments or keyword arguments
                    return valid_calls[node.func.id](node.args)
                else:
                    _raise_malformed_node(node)
            else:
                _raise_malformed_node(node)

        elif isinstance(node, Dict):
            if len(node.keys) != len(node.values):
                _raise_malformed_node(node)
            return dict(zip(map(_convert, node.keys), map(_convert, node.values)))
        elif isinstance(node, BinOp):  # Handle binary operations
            left = _convert(node.left)
            right = _convert(node.right)
            if isinstance(node.op, Add):
                return left + right
            elif isinstance(node.op, Sub):
                return left - right
            elif isinstance(node.op, Mult):
                return left * right
            elif isinstance(node.op, Div):
                return left / right
            elif isinstance(node.op, Mod):
                return left % right
            elif isinstance(node.op, FloorDiv):
                return left // right
            else:
                _raise_malformed_node(node)
        elif isinstance(node, UnaryOp):  # Handle unary operations
            operand = _convert(node.operand)
            # TODO check numeric?
            if isinstance(node.op, UAdd):
                return +operand
            elif isinstance(node.op, USub):
                return -operand
            else:
                _raise_malformed_node(node)
        elif isinstance(node, Subscript):  # Handle indexing operations
            value = _convert(node.value)
            index = _convert(node.slice)
            return value[index]
        else:
            _raise_malformed_node(node)

    try:
        if isinstance(node_or_string, str):
            node_or_string = parse(node_or_string.lstrip(" \t"), mode="eval")
        if isinstance(node_or_string, Expression):
            node_or_string = node_or_string.body
    except SyntaxError:
        _raise_malformed_node(node_or_string)

    return _convert(node_or_string)



"""Module defining the `JoyStickEvent` class (NOT IMPLEMENTED)."""

from ..event import Event


# TODO joystick support may be provided at some point
class JoyStickEvent(Event):
    """A class representing a joystick event (NOT IMPLEMENTED)."""

    pass



"""Package defines UI related functionality. This functionality may be used by `star_ray` extensions to form a common public API."""

from .window_configuration import WindowConfiguration

__all__ = ("WindowConfiguration",)



"""Module defining common exception types."""

from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..event import Event

__all__ = ("UnknownEventType",)


class UnknownEventType(TypeError):
    """Exception raised when an unknown event type is encountered."""

    def __init__(self, event: Event):
        """Constructor.

        Args:
            event (_type_): _description_
        """
        event_type = type(event)
        super().__init__(f"Unknown event type: {event_type}.")
        self.event_type = event_type



"""Module defines the `AgentRouted` class and the `observe` and `decide` decorators. See class documentation for details."""

from typing import Any
from collections.abc import Callable
from functools import wraps
import inspect

from .component import Component
from .component import Sensor, Actuator
from .agent import Agent

from ..utils import TypeRouter
from ..event import Event

__all__ = ("AgentRouted", "observe", "decide")


def _get_decide_methods(obj: Any) -> list[Callable]:
    """Internal method that will retreive all (bound) methods on `obj` that have been tagged with `is_decide`. This tag is added if the method has been decorted with `@decide`.

    Args:
        obj (Any): object to get decide methods from.

    """
    methods = inspect.getmembers(obj, predicate=inspect.ismethod)
    return [m[1] for m in filter(lambda x: getattr(x[1], "is_decide", False), methods)]


def _component_arg(func: Callable) -> Callable:
    """Internal decorate that is used to resolve the optional `component` argument in a method decorated with `@observe`. Observe methods are not required to define the `component` argument, but if it is define the sensor or actuator will be provided.

    Args:
        func (Callable): the method

    Returns:
        Callable: the wrapped method that handles `component` if required, otherwise the original `func` argument.
    """
    kwargs = list(inspect.signature(func).parameters.keys())
    if "component" in kwargs:
        return func
    else:

        @wraps(func)
        def _wrapped(*args, **kwargs):
            del kwargs["component"]
            return func(*args, **kwargs)

        return _wrapped


def decide(*fun: Callable):
    """A decorator that may be added to methods in a `RoutedAgent` that are intended to produce actions for execution (i.e. the decision of the agent). These actions will be automatically routed to components that define type-compatible attempt methods.

    Example:
    ```
    class MyAgent(RoutedAgent):
        @decide
        def my_decide(self) -> Event:
            return MyDecision()  # send to relevant actuators during `__cycle__`


    class MyActuator(Actuator):
        @attempt
        def my_attempt(self, decision: MyDecision) -> MyAction:
            return MyAction()  # the action is now scheduled for execution
    ```

    Note that if multiple compatible components are discovered then all of them will receive the decision/action. Some care may be need to avoid duplicate execution if the decision is an executable action.

    As a principle, attempt methods should always be conservative with their typing if used in routing. There are very few situations where multiple components should define attempt methods that have very generic typings (e.g. `Event` or `Action`) as this can easily lead to duplicate actions being executed, see `@attempt` for further details.
    """

    def _decide(f):
        f.is_decide = True
        return f

    if len(fun) == 0:
        return _decide
    elif len(fun) == 1:
        return _decide(fun[0])
    else:
        raise ValueError("`decide` decorator takes no arguments.")


def observe(*event_types: tuple[type]):
    """A decorator that may be added to methods in `AgentRouted` that are intended to process observations. Such methods will automatically receive type-compatible observations from the agents components (sensors and actuators).

    Args:
        event_types (tuple[type], optional): the types to receive, a single positional argument may be provide i.e. @observe([A,B,C]), if not argument is provided i.e. usage `@observe` or `@observe()` then types will be derived from the type hint of the first argument of the decorated method.

    Example:
    ```
    class MyAgent(RoutedAgent):
        @observe
        def revise_beliefs(self, observation: MyObservation, component: Component):
            # `component` is sensor or actuator that produced the observation, it is optional
            pass  # revise agents beliefs
    ```

    The `component` argument in an observe method is optional a valid signature would also be: `revise_beliefs(self, observation : MyObservation)`

    Raises:
        ValueError: if `event_types` is not specified correctly or multiple positional arguments are provided.
    """

    def _observe(func, route_types=None):
        if route_types is None:
            route_types = TypeRouter.resolve_first_argument_types(
                func, allow_no_arguments=False
            )

        # used to identify whether a given method has been decorated with this decorator.
        func.is_observe = True
        # this is used to automatically route events to a given component.
        if route_types:
            route_types = TypeRouter.validate_types(route_types)
            func.route_types = route_types
        return func

    if len(event_types) == 0:
        return _observe
    elif len(event_types) == 1 and isinstance(event_types[0], list | tuple):

        def _wrap(func, route_types=event_types[0]):
            return _observe(func, route_types=route_types)

        return _wrap
    elif len(event_types) == 1:
        # event_types is actually the fun we are wrapping... used the decorator as @observe without arguments
        return _observe(event_types[0], route_types=None)
    else:
        raise ValueError(
            "Decorator `observe` takes a single positional argument `event_types`."
        )


class AgentRouted(Agent):
    """An agent base class adds automatic routing of observation/decision/actions to relevant methods, sensors and actuators by type. The `__cycle__` method has been defined to handle this.

    # Observe, Decide, Attempt

    Three decorators are used to enable this routing.

    - `observe`: used to route observations from sensors to methods in the agent.
    - `decide` : used to route decisions (or actions) from the agent to type-compatible actuators (specifically their `attempt` methods).
    - `attempt` : used in actuators and sensors as a destination for routed decisions/actions (originating in decide methods).

    `observe` and `decide` are used to decorate methods in the agent.
    `attempt` is used to decorate methods in the agents components (sensors and actuators), recall that it is also used to schedule action for execution during `__sense__` (for sensors) and `__execute__` (for actuators).

    Methods decorated with any of the three are called during `__cycle__`.

    Example:
    ```
    class MyAgent(RoutedAgent):
        @observe
        def my_revise(self, observation: MyObservation, component: Component):
            pass  # observations of type `MyObservation` will be sent here from sensors (or actuators)

        @decide
        def my_decide(self) -> MyDecision:
            return MyDecision()  # send to the type-compatible actuator

        @decide
        def my_sense(self) -> MySense:
            return MySense()  # send to the type-compatible sensor


    class MyActuator(Actuator):
        @attempt
        def my_attempt(self, decision: MyDecision) -> MyAction:
            return MyAction()


    class MySensor(Actuator):
        @attempt
        def my_attempt(self, decision: MySense) -> MyAction:
            return MySenseAction()
    ```
    For more details on each decorator, see their respective documentation.

    Methods that have been decorated with `observe` and `decide` are automatically discovered, but it is also possible to manually add functions via the `add_observe` and `add_decide` methods. The added function does not need to be an instance method.

    During `__cycle__` the following methods dunder methods are called: `__observe__`, `__decide__`.

    - `__observe__` is responsible for routing observations from sensors to type-compatible observe methods.
    - `__decide__` is responsible for routing decisions (or actions) from decide methods to type-compatible actuators (attempt methods).
    """

    _TYPE_ROUTER_CACHE_SIZE = 64  # default value for router cache size

    def __init__(
        self, sensors: list[Sensor], actuators: list[Actuator], *args, **kwargs
    ):
        """Constructor.

        Args:
            sensors (list[Sensor]): initial sensor components.
            actuators (list[Actuator]): initial actuator components.
            args (tuple[Any], optional): optional additional arguments.
            kwargs (dict[str, Any], optional): optional additional keyword arguments.
        """
        # this may be added to in super().__init__ (upon adding an actuator)
        # it will route actions to the type-compatible component attempt methods
        self._attempt_routers: dict[int, TypeRouter] = dict()

        super().__init__(sensors, actuators, *args, **kwargs)
        # router for observations to type-compatible observe methods
        self._observe_router = TypeRouter(
            cache_size=AgentRouted._TYPE_ROUTER_CACHE_SIZE
        )
        # these methods are those that have been decorated with @observe
        # they will recieve observations of the relevant type from the sensors.
        self._observe_methods = list(
            filter(
                lambda m: getattr(m, "is_observe", False),
                TypeRouter.get_all_decorated_methods(self),
            )
        )
        for method in self._observe_methods:
            self._observe_router.add(_component_arg(method))

        # these methods are those that have been decorated with @decide, they are to be called in `__cycle__`.
        self._decide_methods = _get_decide_methods(self)

    def add_observe(self, func: Callable, event_types: list[type[Event]] | None = None):
        """Adds a new observe function. The function need not be decorated with `@observe`.

        Args:
            func (Callable): function to add.
            event_types (list[type], optional): `event_types` to route to the function. Defaults to None. If not specified the types will be resolved from the type hints of the first argument of `func` argument (unless the @observe decorator was used).
        """
        func = _component_arg(func)  # handle optional "component" argument
        self._observe_methods.append(func)
        self._observe_router.add(func, event_types)

    def add_decide(self, func: Callable):
        """Adds a new decide function. The function need not be decorated with `@decide`.

        Args:
            func (Callable): function to add.
        """
        if not hasattr(func, "is_decide"):
            func = decide(func)
        self._decide_methods(func)

    def attempt(self, action: Event | list[Event], component: Component | None = None):
        """Manually attempt one or more actions and route to the provide component.

        If `component` is not provided, this method will search for the relevant component(s) to attempt the `action` based on its type using the usual action/decision routing mechanism.

        Args:
            action (Event | List[Event]): action (or actions) to attempt.
            component (Component, optional): to attempt with. Defaults to None.
        """
        if not isinstance(action, list | tuple):
            action = (action,)
        if component is None:
            for attempt_router in self._attempt_routers.values():
                for a in filter(None, action):
                    attempt_router(a)
        else:
            attempt_router = self._attempt_routers[component.id]
            for a in filter(None, action):
                attempt_router(a)

    def add_component(self, component: Component) -> Component:  # noqa
        component = super().add_component(component)
        # get attempt methods from the component (if any) and register them with a new TypeRouter
        attempt_methods = list(
            filter(
                lambda m: getattr(m, "is_attempt", False),
                TypeRouter.get_all_decorated_methods(component),
            )
        )
        if len(attempt_methods) > 0:
            attempt_router = TypeRouter(cache_size=AgentRouted._TYPE_ROUTER_CACHE_SIZE)
            for method in attempt_methods:
                attempt_router.add(method)
            self._attempt_routers[component.id] = attempt_router
        else:
            pass  # no attempt methods were defined
        return component

    def remove_component(self, component: Component) -> Component:  # noqa
        component = super().remove_component(component)
        if component.id in self._attempt_routers:
            del self._attempt_routers[component.id]
        return component

    def __cycle__(self):  # noqa
        self.__observe__()
        self.__decide__()

    def __decide__(self) -> None:
        """Method that routes decisions/actions to type-compatible component attempt methods. Called once per cycle."""
        resulting_actions = []
        for method in self._decide_methods:
            # the method is already bound
            decisions = method()
            if not isinstance(decisions, list | tuple):
                decisions = [decisions]
            # forward the actions to the relevant actuators
            for decision in filter(None, decisions):
                for attempt_router in self._attempt_routers.values():
                    action = attempt_router(decision)
                    resulting_actions.append(action)
        return resulting_actions  # TODO do something with these?

    def __observe__(self):
        """Method that routes observations from components to type-compatible observe methods. Called once per cycle."""
        for actuator in self.get_actuators():
            for observation in actuator.iter_observations():
                assert isinstance(observation, Event)
                self._observe_router(observation, component=actuator)
        for sensor in self.get_sensors():
            for observation in sensor.iter_observations():
                assert isinstance(observation, Event)
                self._observe_router(observation, component=sensor)



# TODO check if this is used anywhere? refactor by removing!
from typing import Any

import ray

__all__ = ("new_instance_local_or_remote",)

_actor_class_cache: dict[type, ray.actor.ActorClass] = {}


def new_instance_local_or_remote(
    cls: type,
    args: list[Any] = None,
    kwargs: dict[str, Any] = None,
    remote: bool = False,
) -> Any | ray.actor.ActorHandle:
    """Dynamically create a local instance or a [ray] remote actor based on the given class.

    Args:
        cls ([Type]): The class to instantiate.
        args ([List[Any]], optional): Positional arguments to pass to the class constructor.
        kwargs ([Dict[str, Any]], optional): Keyword arguments to pass to the class constructor.
        remote ([bool], optional): Whether to create a Ray remote actor. Defaults to False.

    Returns:
        [Union[Any, ray.actor.ActorHandle]]: An instance of `cls` or a Ray remote actor handle for `cls`.
    """
    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}

    if remote:
        if cls not in _actor_class_cache:
            _actor_class_cache[cls] = ray.remote(cls)
        ActorClass = _actor_class_cache[cls]
        return ActorClass.remote(*args, **kwargs)
    else:
        return cls(*args, **kwargs)



"""Internal base classes that manage local/remote/async agents in an `Ambient`. They make use of the `_Future` API to streamline method/property access."""

from __future__ import annotations
from typing import Any, TYPE_CHECKING
from abc import ABC, abstractmethod
import asyncio
import ray
from ray.actor import ActorHandle

from ..utils import _Future
from .agent import Agent

if TYPE_CHECKING:
    from ..environment import State

__all__ = ("_Agent",)


class _Agent(ABC):
    """Internal base class for streamlining method/property access on agents for local/remote/async agents."""

    def __init__(self, agent: Any):
        super().__init__()
        self._inner = agent

    @staticmethod
    def is_agent_async(agent: Any):
        """Does the agent define asynchronous methods? (declared `async`)."""
        return (
            asyncio.iscoroutinefunction(agent.__sense__)
            and asyncio.iscoroutinefunction(agent.__execute__)
            and asyncio.iscoroutinefunction(agent.__cycle__)
        )

    @staticmethod
    def is_agent_sync(agent: Any):
        """Does the agent define synchronous methods? (not declared `async`)."""
        return not (
            asyncio.iscoroutinefunction(agent.__sense__)
            or asyncio.iscoroutinefunction(agent.__execute__)
            or asyncio.iscoroutinefunction(agent.__cycle__)
        )

    @staticmethod
    def is_agent_remote(agent: Any):
        """Is the agent a remote ray actor?"""
        return isinstance(agent, ActorHandle)

    @staticmethod
    def new(agent: Any):
        """Factory function that will wrap the given agent in one of three wrappers: `_AgentWrapperRemote`, `_AgentWrapperLocal`, `_AgentWrapperAsync` depending on its type and whether its method definitions are `async`.

        Args:
            agent (Any): the agent (or remote handle) to wrap.

        Raises:
            TypeError: If the agent does not meet the type requirements.

        Returns:
            _Agent: the wrapped agent.
        """
        if isinstance(agent, _Agent):
            return agent
        elif _Agent.is_agent_remote(agent):
            return _AgentWrapperRemote(agent)
        elif isinstance(agent, Agent):
            # the agent is local, is it sync or async?
            if _Agent.is_agent_async(agent):
                return _AgentWrapperLocalAsync(agent)
            elif _Agent.is_agent_sync(agent):
                return _AgentWrapperLocal(agent)
            else:
                raise TypeError(
                    f"Invalid method definitions in agent {agent}, __cycle__, __sense__, __execute__ must be declared all async or all sync."
                )
        raise TypeError(f"Invalid type for agent {type(agent)}.")

    @abstractmethod
    def __initialise__(self, state: State) -> _Future:
        pass

    @abstractmethod
    def __sense__(self, state: State) -> _Future:
        pass

    @abstractmethod
    def __cycle__(self) -> _Future:
        pass

    @abstractmethod
    def __execute__(self, state: State) -> _Future:
        pass

    @abstractmethod
    def __terminate__(self) -> _Future:
        pass

    @abstractmethod
    def get_id(self) -> Any:
        pass

    @abstractmethod
    def get_inner(self) -> Any:
        pass

    @property
    @abstractmethod
    def is_alive(self) -> bool:
        pass

    @property
    @abstractmethod
    def id(self) -> Any:
        pass


class _AgentWrapperRemote(_Agent):
    def __initialise__(self, state: State) -> _Future:
        return _Future.call_remote(self._inner.__initialise__, state)

    def __sense__(self, state: State) -> _Future:
        return _Future.call_remote(self._inner.__sense__, state)

    def __cycle__(self) -> _Future:
        return _Future.call_remote(self._inner.__cycle__)

    def __execute__(self, state: State) -> _Future:
        return _Future.call_remote(self._inner.__execute__, state)

    def __terminate__(self) -> _Future:
        return _Future.call_remote(self._inner.__terminate__)
        # TODO?
        # ray.kill(self._inner, no_restart=True)

    def get_inner(self):
        return self._inner

    def get_id(self):
        return self.id

    @property
    def is_alive(self) -> bool:
        return ray.get(self._inner.get_is_alive.remote())

    @property
    def id(self) -> Any:
        return ray.get(self._inner.get_id.remote())


class _AgentWrapperLocal(_Agent):
    def __initialise__(self, state: State) -> _Future:
        return _Future.call_sync(self._inner.__initialise__, state)

    def __sense__(self, state: State) -> _Future:
        return _Future.call_sync(self._inner.__sense__, state)

    def __cycle__(self) -> _Future:
        return _Future.call_sync(self._inner.__cycle__)

    def __execute__(self, state: State) -> _Future:
        return _Future.call_sync(self._inner.__execute__, state)

    def __terminate__(self):
        return _Future.call_sync(self._inner.__terminate__)

    def get_inner(self):
        return self._inner

    def get_id(self):
        return self._inner.id

    @property
    def is_alive(self) -> bool:
        return self._inner.is_alive

    @property
    def id(self) -> Any:
        return self._inner.id


class _AgentWrapperLocalAsync(_AgentWrapperLocal):
    def __initialise__(self, state: State) -> _Future:
        return _Future.call_async(self._inner.__initialise__, state)

    def __sense__(self, state: State) -> _Future:
        return _Future.call_async(self._inner.__sense__, state)

    def __cycle__(self) -> _Future:
        return _Future.call_async(self._inner.__cycle__)

    def __execute__(self, state: State) -> _Future:
        return _Future.call_async(self._inner.__execute__, state)

    def __terminate__(self):
        return _Future.call_async(self._inner.__terminate__)



"""Module defines the `KillEvent` class, which is used to terminate an agent. See class documentation for details."""

from pydantic import model_validator, Field
from typing import ClassVar
from .event import Event
from ..utils import int64_uuid


# TODO untested
class KillEvent(Event):
    """Event that will kill an agent given its `id` (trigger a call to __terminate__). It is up to the environment to manage the permissions of this event.

    Attributes:
        kill (int, optional): The id of the agent to kill. Defaults the source of this event (if executed by an agent, the agent will kill itself).
    """

    kill: int = Field(default_factory=lambda: KillEvent._DEFAULT_UUID)

    @model_validator("kill", mode="after")
    def _validate_kill(self):
        if self.kill == KillEvent._DEFAULT_UUID:
            self.kill = self.source

    _DEFAULT_UUID: ClassVar[int] = int64_uuid()



"""Module defining the `Actuator` class. See class documentation for details."""

from __future__ import annotations
from typing import TYPE_CHECKING

from .component import Component

if TYPE_CHECKING:
    from ...environment import State

__all__ = ("Actuator",)


class Actuator(Component):
    """Base class for an actuator. `Actuator`s may be attached to an `Agent` and will allow them to take actions to mutate the environment state. They run as part of `__execute__` in the agents cycle.

    Like sensors, actuators are typically defined with a number of "attempt" methods. These methods are decorated with the `@attempt` decorator. The decorator ensures that the action is prepared for execution after the agent has decided upon it.

    A typical example might look as follows:
    ```
    class MyAgent(Agent):
        def __cycle__(self):
            self.my_actuator.move(self.direction)


    class MyActuator(Actuator):
        @attempt()
        def move(self, direction: float):
            return MoveAction(direction)
    ```

    An attempt method will always return the action that the agent wishes to take. It is possible to directly pass an action to an attempt method, in which case the method might simply return it:
    ```
    @attempt()
    def take(self, action: Action):
        return action
    ```
    More often, there will be some additional steps inside the attempt method which prepares the action for execution. This simplifies the logic present inside the agents `__cycle__` method and can allow for some more generic setups where the agent is unaware of the implementation details of an action.

    In some situations we may want an actuator to always take some action (or set of actions) regardless of the agents decision. In this case we can override the `__attempt__` method, see method documentation for details.

    Actuators can also attached to agents that subclass `AgentRouted`, which will allow for automatic routing of decisions to their respective actuators (we don't need to call attempt methods explicitly) see this class for details.
    """

    def __query__(self, state: State) -> None:
        """Method that will execute all current actions in the given environment `state`.

        Args:
            state (State): to execute actions in.

        Raises:
            TypeError: if `__attempt__` does not return the correct type.
        """
        # get actions
        actions = self.__attempt__()
        if not isinstance(actions, list | tuple):
            raise TypeError(
                f"`__attempt__` must return a `tuple` of actions, received: {actions}"
            )
        self._actions.extend(actions)
        self._actions = list(filter(None, self._actions))
        # set the source of these actions to this actuator
        Component.set_action_source(self, self._actions)
        # attempt the sense act and get the resulting observations
        observations = state.__update__(self._actions)
        # preprocess the observations ready to be received by the agent
        self._observations.push_all(self.__transduce__(observations))
        # clear sense actions ready for the next execution cycle
        self._actions.clear()

    def __attempt__(self):
        """This method can be overriden to create an actuator that does not depend on calls to `attempt` methods. This is useful if the actuator should always perform the same kind of action independently of the agents decision making.

        The default implementation of this method simply returns any actions returned by an `attempt` methods.

        Example:
        ```
        class FowardActuator(Actuator):
            def __attempt__(self):
                # new actions should always be created, do not re-use events!
                return [QueryMoveForward()]
        ```
        See also:
            [`Sensor.__sense__`]
        """
        return []



"""Module defines a utility class `WindowConfiguration` that may be used to configure application windows."""

from pydantic import BaseModel, Field


class WindowConfiguration(BaseModel):
    """Configuration options for an application window."""

    x: int = Field(default=0)
    y: int = Field(default=0)
    width: int = Field(default=640)
    height: int = Field(default=480)
    title: str = Field(default="window")
    resizable: bool = Field(default=False)
    fullscreen: bool = Field(default=False)
    background_color: str = "#ffffff"



"""Internal base classes that manage local/remote observations in `Component` implementations."""

from typing import Any
from abc import ABC, abstractmethod
from collections import deque
import ray

from ..event import Event


class _Observations(ABC):
    """Abstract base class for managing collections of observations, either remote or local. This class provides a common interface for operations on a collection of observations, which can be either local (`Event` objects) or remote (`ray.ObjectRef` objects).

    The class is used internally by `Component` classes to seemlessly manage the retreival of remote (or local) data.

    Important note: Iterating over the collection will consume it.
    """

    @staticmethod
    def new(objects: list[Event | ray.ObjectRef]):
        if len(objects) == 0:
            return _Observations.empty()
        elif isinstance(objects[0], ray.ObjectRef):
            return _ObservationsRemote(objects)
        elif isinstance(objects[0], Event):
            return _ObservationsLocal(objects)
        else:
            raise TypeError(f"Invalid observation type {type(objects[0])}.")

    @staticmethod
    def empty():
        return _ObservationsLocal([])

    @abstractmethod
    def is_empty(self):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def push_all(self, items: list[Any]) -> None:
        pass

    @abstractmethod
    def push(self, item: Any) -> None:
        pass

    @abstractmethod
    def pop(self) -> Event:
        pass

    @abstractmethod
    def __iter__(self):
        pass

    @abstractmethod
    def __next__(self) -> Event:
        pass


class _ObservationsRemote(_Observations):
    """Collection of remote observations."""

    def __init__(self, objects):
        super().__init__()
        self._objects = deque(objects)

    def __len__(self):
        return self._objects.count

    def is_empty(self):
        return self._objects.count == 0

    def push(self, item: ray.ObjectRef) -> None:
        self._objects.append(item)

    def push_all(self, items: list[ray.ObjectRef]) -> None:
        self._objects.extend(items)

    def pop(self) -> Event:
        return ray.get(self._objects.popleft())

    def __iter__(self):
        return self

    def __next__(self) -> Event:
        if not self._objects:
            raise StopIteration
        return ray.get(self._objects.popleft())


class _ObservationsLocal(_Observations):
    """Collection of local observations."""

    def __init__(self, objects):
        super().__init__()
        self._objects = deque(objects)

    def __len__(self):
        return self._objects.count

    def is_empty(self):
        return self._objects.count == 0

    def push_all(self, items: list[Event]) -> None:
        self._objects.extend(items)

    def push(self, item: Event) -> None:
        self._objects.append(item)

    def pop(self) -> Event:
        return self._objects.popleft()

    def __iter__(self):
        return self

    def __next__(self) -> Event:
        if not self._objects:
            raise StopIteration
        return self._objects.popleft()



"""Package defining pub-sub related functionality.

TODO describe pubsub, important classes and give an example!
"""

from ._action import Subscribe, Unsubscribe
from ._pubsub import Subscriber, Publisher
from ._topic_publisher import TopicPublisher
from ._type_publisher import TypePublisher

__all__ = (
    "Subscribe",
    "Unsubscribe",
    "Subscriber",
    "Publisher",
    "TopicPublisher",
    "TypePublisher",
)



"""Module defining the `MouseButtonEvent` and `MouseMotionEvent` classes."""

from typing import ClassVar
from pydantic import field_validator, Field
from ..event import Event

__all__ = ("MouseButtonEvent", "MouseMotionEvent")


class MouseButtonEvent(Event):
    """A class representing a mouse event.

    Attributes:
        id ([int]): A unique identifier for the event.
        timestamp ([float]): The timestamp (in seconds since UNIX epoch) when the event instance is created.
        source ([int]): A unique identifier for the source of this event.
        button ([int]): The mouse button involved in the event (1 for left click, 2 for middle click, 3 for right click, etc.).
        position ([tuple]): The (x, y) coordinates of the mouse event.
        status ([int]): The status of the event (UP = 0, DOWN = 1, CLICK = 2).
        target ([str]): The UI element that was clicked on. This value is UI implementation dependent and may be None, typically it will be a unique element ID.
    """

    button: int
    position: tuple[float, float] | tuple[int, int]
    status: int
    target: str | list[str] | None = Field(default=None)

    # static fields
    UP: ClassVar[int] = 0
    DOWN: ClassVar[int] = 1
    CLICK: ClassVar[int] = 2

    BUTTON_LEFT: ClassVar[int] = 0
    BUTTON_MIDDLE: ClassVar[int] = 1
    BUTTON_RIGHT: ClassVar[int] = 2

    @field_validator("position", mode="before")
    @classmethod
    def _validate_position(cls, value: tuple | dict | list):
        if isinstance(value, dict):
            return (value["x"], value["y"])
        else:
            return tuple(value)

    @field_validator("status", mode="before")
    @classmethod
    def _validate_status(cls, value: str | int) -> int:
        if isinstance(value, str):
            if "release" in value or value == "up":
                return MouseButtonEvent.UP
            elif "press" in value or value == "down":
                return MouseButtonEvent.DOWN
            elif "click" in value:
                return MouseButtonEvent.CLICK
            else:
                raise ValueError(
                    f"Failed to convert status string: {value}, valid values include: ['released', 'pressed', 'clicked', 'up', 'down']"
                )
        elif isinstance(value, int) and value in [
            MouseButtonEvent.UP,
            MouseButtonEvent.DOWN,
            MouseButtonEvent.CLICK,
        ]:
            return value
        else:
            raise ValueError(f"Invalid {type(MouseButtonEvent)} status code: {value}")


class MouseMotionEvent(Event):
    """A class representing a mouse motion event.

    Attributes:
        id ([int]): A unique identifier for the event.
        timestamp ([float]): The timestamp (in seconds since UNIX epoch) when the event instance is created.
        source ([int]): A unique identifier for the source of this event.
        position: ([Tuple[float,float]]): The window relative position of the mouse pointer.
        relative ([Tuple[float,float]]): The relative motion of the mouse since the last event.
        target ([str | List[str]]): The UI element that the mouse is currently over. This value is UI implementation dependent and may be None, typically it will be a unique element ID.
    """

    position: tuple[float, float] | tuple[int, int]
    relative: tuple[float, float] | tuple[int, int]
    target: str | list[str] | None = Field(default=None)



"""Module defining the `Sensor` class. See class documentation for details."""

from __future__ import annotations
from typing import TYPE_CHECKING

from .component import Component
from ...event import Event
from ...pubsub import Subscriber, Subscribe, Unsubscribe

if TYPE_CHECKING:
    from ..agent import Agent
    from ...environment import State

__all__ = ("Sensor",)


class Sensor(Component, Subscriber):
    """Base class for a sensor. `Sensors`s may be attached to an `Agent` and will allow them to "sense" the current state of the environment. A sensor does this by taking "sense actions", which are read only actions. sense actions occur as part of `__sense__` in the agents cycle.

    Sensing often happens without an explicit decision made by the agent. The `__sense__` method can be overriden to continuously (once per cycle) take a set of sense actions.

    __Example__:
    ```
    class MySensor(Sensor):
        def __sense__(self):
            # new actions should always be created, do not re-use events!
            return [GetPositionAction()]
    ```

    These observations are then avaliable in the agents cycle via the `iter_observations` method.

    Example:
    ```
    class MyAgent(Agent):
        def __cycle__(self):
            for observation in self.my_sensor.iter_observations():
                print(observation)  # update beliefs or make decisions
    ```
    Note that `iter_observations` will consume the observations, more will not be avaliable until the next cycle (after the next round of sensing).

    Sensors may also take advance of `star_ray`s pub-sub mechanism (`Sensor` implements the `Subscriber` interface) and can subscribe to passively receive certain types of event. The `__subscribe__` method may be used for this purpose, it is called once automatically upon initialisation of the sensor (after it is has been attached to an agent) and should return a list of `Subscribe` actions. The sensors observations will then automatically be populated with any published data that it subscribed to. Note that it is up to the environment to define what data is made avaliable to a sensor (i.e. what `Publisher`s exist).

    Like actuators, sensors may be defined with a number of "attempt" methods. These methods are decorated with the `@attempt` decorator. The decorator ensures that the action is prepared for execution after the agent has decided upon it. Since sensing happens __before__ the call to `__cycle__` (see `Agent` class documentation for details), decisions to actively sense are made on the previous cycle.

    A typical example might look as follows:
    ```
    class MyAgent(Agent):
        def __cycle__(self):
            self.my_sensor.move(self.direction)


    class MySensor(Sensor):
        @attempt()
        def get_position(self):
            # this action is defined according to the environments state API.
            return GetPositionAction()
    ```

    An attempt method will always return the action that the agent wishes to take. It is possible to directly pass an action to an attempt method, in which case the method might simply return it:
    ```
    @attempt()
    def sense(self, action: Action):
        return action
    ```
    More often, there will be some additional steps inside the attempt method which prepares the action for execution. This simplifies the logic present inside the agents `__cycle__` method and can allow for some more generic setups where the agent is unaware of the implementation details of an action.

    Sensors can also attached to agents that subclass `AgentRouted`, which will allow for automatic routing of received observations to methods defined in the agent that are decorated with `@observe`. As with actuators, the same kind of action routing can be done (see `Actuator` class documentation for details).
    """

    def __query__(self, state: State) -> None:
        """Method that will execute all current actions in the given environment `state`.

        Args:
            state (State): to execute actions in.

        Raises:
            TypeError: if `__sense__` does not return the correct type.
        """
        # get sense actions
        actions = self.__sense__()
        if not isinstance(actions, list | tuple):
            raise TypeError(
                f"`__sense__` must return a `tuple` of actions, received: {actions}"
            )
        self._actions.extend(actions)
        self._actions = list(filter(None, self._actions))

        # set the source of these actions to this sensor
        Component.set_action_source(self, self._actions)
        # attempt the sense act and get the resulting observations
        observations = state.__select__(self._actions)
        # preprocess the observations ready to be received by the agent
        self._observations.push_all(self.__transduce__(observations))
        # clear sense actions ready for the next execution cycle
        self._actions.clear()

    def __notify__(self, message: Event) -> None:
        """Called by a `Publisher` with events that the sensor has subscribed to receive.

        Args:
            message (Event): the event that sensor has subscribed to receive.
        """
        self._observations.push_all(self.__transduce__([message]))

    def __subscribe__(self) -> list[Subscribe]:
        """Subscription actions for this `Sensor`, these actions will be taken on the first call to `__query__`. During the lifetime of this `Sensor` subscription actions can be taken in the usual way as with any other action.

        Returns:
            List[Subscribe]: list of `Subscribe` actions to take.

        Example:
            TODO
        """
        return []

    def on_add(self, agent: Agent) -> None:  # noqa: D102
        super().on_add(agent)
        # call __subscribe__ to take initial subscription actions (if there are any)
        self._actions.extend(self.__subscribe__())
        Sensor.set_subscriber(self, self._actions)

    def __sense__(self) -> list[Event]:
        """This method can be overridden to create a `Sensor` that does not depend on calls to `attempt` methods. This is useful if the `Sensor` should always sense the same kind of data.

        __Example__:
        ```
        class MySensor(Sensor):
            def __sense__(self):
                # new actions should always be created, do not re-use events!
                return [GetPositionAction()]
        ```

        See Also:
            [`Actuator.__attempt__`]
        """
        return []

    @staticmethod
    def set_subscriber(sensor: Sensor, actions: list[Subscribe | Unsubscribe]):
        """Sets the subscriber to the given sensor sensor if it is None for each action in actions.

        Args:
            sensor (Sensor): sensor to use as the subscriber.
            actions (List[Subscribe  |  Unsubscribe]): action for which to set `subscriber` attribute.
        """
        for action in actions:
            if action.subscriber is None:
                # TODO this will currently call __notify__ of the agent... not this sensor!
                # it only works locally at the moment
                action.subscriber = sensor



"""Module containing utilities for dictionaries."""

from typing import Any


def merge_nested(d1: dict[Any, Any], d2: dict[Any, Any]) -> dict[Any, Any]:
    """Recursively merges two nested dictionaries. Only merges the leaves from d2 into d1.

    Args:
        d1 (dict): The target dictionary into which `d2` is merged. This dictionary is modified in place.
        d2 (dict): The source dictionary from which key-value pairs are taken. `d2` remains unmodified.

    Returns:
        dict: The updated dictionary `d1` after merging in `d2`. Note that `d1` is modified in place, but also returned for convenience.

    Example:
        ```
        dict1 = {"a": 1, "b": {"x": 2, "y": 3}}
        dict2 = {"b": {"y": 4, "z": 5}, "c": 3, "d": {"e": 6}}
        merged_dict = merge_nested_dicts(dict1, dict2)
        print(merged_dict)
        {"a": 1, "b": {"x": 2, "y": 4, "z": 5}, "c": 3, "d": {"e": 6}}
        ```

    """
    for key in d2:
        if key in d1:
            if isinstance(d1[key], dict) and isinstance(d2[key], dict):
                merge_nested(d1[key], d2[key])
            else:
                # If the key exists in both but is not a dict in either,
                # or it's a dict in one but not the other, d2's value takes precedence.
                d1[key] = d2[key]
        else:
            # If the key is not in d1, add it from d2.
            d1[key] = d2[key]
    return d1



"""Package that defines component related functionality.

`Component`s can be attached to agents allowing them to perform certain actions and receive observations. They are in essence how the agent interfaces, or "interacts" with its enviroment.

Two important base classes:
    `Sensor` : TODO
    `Actuator` : TODO

This module also defines important decorators that may be used by an `Agent` to route actions/observations to methods that it defines. This can simplify how actions/observations are taken and processed during the agent's cycle.

- `@observe` : TODO
- `@decide` : TODO
- `@attempt` : TODO

See associated documentation for further details.

"""

from .component import Component
from .actuator import Actuator
from .attempt import attempt
from .on_awake import OnAwake
from .sensor import Sensor
from .sensor_io import IOSensor

__all__ = (
    "OnAwake",
    "attempt",
    "observe",
    "Component",
    "Sensor",
    "Actuator",
    "IOSensor",
)



"""Module implements UUID functionality that is used internally to generate unique ids for various objects (including events, agents and ambients)."""

import time
import random
import uuid


def int64_uuid() -> int:
    timestamp = int(time.time() * 1000)  # Convert current time to milliseconds
    rand_number = random.getrandbits(32)  # Generate a 32-bit random number
    unique_id = (timestamp << 32) | rand_number  # Combine timestamp and random number
    return unique_id


def str_uuid4() -> str:
    return str(uuid.uuid4())


__all__ = ("int64_uuid", "str_uuid4")



"""star-ray (Simulation Test-bed for Agent Research)-ray package."""

from .environment import Environment, Ambient
from .agent import Agent, Component, Actuator, Sensor, attempt, observe, decide
from .event import Event

__all__ = (
    "Environment",
    "Ambient",
    "Agent",
    "attempt",
    "observe",
    "decide",
    "Component",
    "Actuator",
    "Sensor",
    "Event",
)



"""Module defines the `_Future` class which is used internally to abstract away from differences between local, remote and async functions/classes."""

import asyncio
import ray
from ray.remote_function import RemoteFunction
from typing import Any
from collections.abc import Callable


__all__ = ("_Future",)


class _Future:
    """A wrapper class that provides a unified interface for asyncio futures, tasks, and Ray object references. This class makes it easier to work with asynchronous and distributed computations in a streamlined/consistent manner, abstracting away the differences between asyncio and Ray."""

    def __init__(self, future: asyncio.Future | ray.ObjectRef):
        super().__init__()
        self._future = future

    @staticmethod
    async def gather(futures: list["_Future"]):
        return await asyncio.gather(*(future.__await__() for future in futures))

    @staticmethod
    def call_sync(func: Callable[..., Any], *args, **kwargs):
        asyncio_future = asyncio.Future()
        try:
            result = func(*args, **kwargs)
            asyncio_future.set_result(result)
        except Exception as e:
            asyncio_future.set_exception(e)
        return _Future(asyncio_future)

    @staticmethod
    def call_async(func: Callable[..., Any], *args, **kwargs):
        # TODO we need to handle exceptions properly here...?
        return _Future(asyncio.create_task(func(*args, **kwargs)))

    @staticmethod
    def call_remote(func: Callable[..., Any], *args, **kwargs):
        return _Future(func.remote(*args, **kwargs))

    @staticmethod
    def wrap(func: Callable[..., Any]):
        """Wraps a given function to ensure that it returns an instance of _Future when called. This allows for a unified interface when working with async, sync, and Ray remote functions. The returned function, when called, will automatically wrap its result in a _Future object.

        Args:
            func (Callable[..., Any]): The function to wrap. This can be an asynchronous
                coroutine function, a synchronous function, or a Ray remote function.

        Returns:
            Callable[..., _Future]: A function that, when called, returns an instance of
            _Future wrapping the result of the original function call.

        Raises:
            TypeError: If the provided `func` argument is not callable.
        """
        if isinstance(func, RemoteFunction):

            def _wrap_remote_with_future(*args, **kwargs) -> "_Future":
                # Wrap the Ray object reference in _Future
                return _Future(func.remote(*args, **kwargs))

            return _wrap_remote_with_future

        elif asyncio.iscoroutinefunction(func):

            def _wrap_with_future(*args, **kwargs) -> "_Future":
                # Wrap the asyncio task in _Future
                return _Future(asyncio.create_task(func(*args, **kwargs)))

            return _wrap_with_future

        elif callable(func):

            def _wrap_sync_with_future(*args, **kwargs) -> "_Future":
                # Immediately resolve the result for a synchronous function
                asyncio_future = asyncio.Future()
                asyncio_future.set_result(func(*args, **kwargs))
                # Wrap the asyncio future in _Future
                return _Future(asyncio_future)

            return _wrap_sync_with_future
        else:
            raise TypeError(f"Failed to wrap: {func}, it is not callable.")

    def __await__(self):
        return (yield from self._future.__await__())


if __name__ == "__main__":
    # example usage
    import time

    START = time.time()

    def sync_fun():
        time.sleep(0.1)
        return f"sync-result-{time.time() - START}"

    async def async_func():
        await asyncio.sleep(0.1)
        return f"async-result-{time.time() - START}"

    @ray.remote
    def remote_func():
        time.sleep(0.1)
        return f"remote-result-{time.time() - START}"

    async def main():
        ray.init()
        future_sync = _Future.call_sync(sync_fun)
        future_async = _Future.call_async(async_func)
        future_remote = _Future.call_remote(remote_func)
        futures = [future_sync, future_async, future_remote]
        results = await _Future.gather(futures)
        print(results)

    asyncio.run(main())



# TODO check if this is used anywhere? refactor by removing!

from typing import Any
import weakref
from collections import UserDict


class DictObservable(UserDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._callbacks = weakref.WeakSet()

    def _trigger_callbacks(self, key: Any, values: tuple[Any, Any]):
        for callback in self._callbacks:
            callback(key, values)

    def __setitem__(self, key: Any, value: Any):
        if key in self:
            old_value = self[key]
            super().__setitem__(key, value)
            self._trigger_callbacks(key, (old_value, value))
        else:
            super().__setitem__(key, value)
            self._trigger_callbacks(key, (None, value))

    def __delitem__(self, key: Any):
        if key in self:
            value = self[key]
            super().__delitem__(key)
            self._trigger_callbacks(key, (value, None))

    def add_callback(self, callback):
        self._callbacks.add(callback)

    def remove_callback(self, callback):
        self._callbacks.remove(callback)


# Example usage:
if __name__ == "__main__":
    x = DictObservable()
    x.add_callback(print)
    x[0] = 1
    x.update({0: 2, 1: 3})
    x.pop(0)

    isinstance(x, dict)



"""Package that defines environment related functionality.

Important classes:
    - `Environment`: the container and entry point of an agent simulation.
    - `Ambient`: defines the state of the environment and holds references to all agents in the simulation.
"""

from .environment import Environment
from .ambient import Ambient, _Ambient

State = _Ambient  # TODO temporary, we need to think more about how the environment state is going to be provided to agents

__all__ = ("Environment", "Ambient", "State")



"""TODO."""

from .component import Component


def OnAwake(cls):
    """A class decorator that will wrap an `Component` class so that it only runs on its first cycle, after which it will remove itself from the `Agent` it is attached to."""
    if not issubclass(cls, Component):
        raise TypeError(
            f"OnAwake decorator can only be applied to classes that derive {Component}."
        )

    class OnAwakeComponent(cls):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._first_cycle = True

        # TODO type hints?
        def __query__(self, state):
            if not self._first_cycle:
                self._agent.remove_component(self)
                return
            super().__query__(state)
            self._first_cycle = False

    return OnAwakeComponent



"""Package that defines agent related functionality.

Important classes:
    - `Agent` : the base class for an agent.
    - `Component` : the base class for a component (see `Actuator` and `Sensor`).
    - `Actuator` : A component that allows the agent to take actions in its environment.
    - `Sensor` : A component that allows an agent to observe its environment.

Other useful classes:
    - `AgentRouted` : An implementation of `Agent` that automatically routes actions/observations to assocaited actuators/sensors (based on types).

See respecitve class documentation for details.
"""

from ._wrapper_agent import _Agent
from ._wrapper_observations import _Observations

from .agent import Agent
from .agent_routed import AgentRouted, decide, observe

from .component import (
    attempt,
    Component,
    Sensor,
    Actuator,
    IOSensor,
)

__all__ = (
    "_Agent",
    "_Observations",
    "Agent",
    "AgentRouted",
    "decide",
    "attempt",
    "observe",
    "Component",
    "Sensor",
    "IOSensor",
    "Actuator",
)



"""Module defines the `Environment` class.

The environment is the container in which the simulation runs and is the simulation entry point. It manages the execution of agents, and has a state (the `Ambient`) which agents read and mutate.
"""

from __future__ import annotations
from typing import TYPE_CHECKING

import asyncio
from .ambient import Ambient, _Ambient
from ..utils import _Future, _LOGGER

if TYPE_CHECKING:
    from ..agent import _Agent


class Environment:
    """The environment is the container in which the simulation runs and is the simulation entry point. It manages the execution of agents, and has a state (the `Ambient`) which agents read and mutate."""

    def __init__(
        self, ambient: Ambient, sync: bool = True, wait: float = 0.05, **kwargs
    ):
        """Constructor.

        Args:
            ambient (Ambient): the state of the environment.
            sync (bool, optional): whether to run the agents synchronously or not. Under the default schedule, if True this means that each cycle method will be gathered together for all agents - i.e. all agents will `__sense__` then `__cycle__` then `__execute__`. If False, then these methods will execute in not particular order, however there will always be a sync point at the start of each cycle.
            wait (float, optional): time to wait between cycles, this leaves room for other async operations if required. Defaults to 0.05.
            **kwargs (dict[str,Any], optional): optional additional arguments.
        """
        super().__init__()
        self._wait = wait
        self._ambient = _Ambient.new(ambient)
        self._step = self._step_sync if sync else self._step_async
        self._cycle = 0

    def run(self):
        """Entry point of the simulation, this call is blocking."""

        async def _run():
            event_loop = asyncio.get_event_loop()
            await self.__initialise__(event_loop)
            _LOGGER.debug("Environment running...")
            await asyncio.gather(*self.get_schedule())

        asyncio.run(_run())

    async def __initialise__(self, event_loop: asyncio.AbstractEventLoop):
        """Initialise this environment. Override this for custom initialisation.

        Args:
            event_loop (AbstractEventLoop): the asyncio event loop that is in use.
        """
        await self._ambient.__initialise__()

    def get_schedule(self) -> list[asyncio.Task]:
        """Get all asyncio tasks that are to run during the simulation. This will include one or more schedulars that manange the execution of the agents, but may also include other environmental processes.

        Returns:
            list[Task]: list of tasks that should be run during the simulation.
        """
        return [asyncio.create_task(self._loop())]

    async def _loop(self):
        """Default schedule."""
        # await self.initialise_agents()
        running = True
        while running:
            running = await self.step()
            await asyncio.sleep(self._wait)
        _LOGGER.debug("--- MAIN SIMULATION LOOP COMPLETED --- ")

    async def step(self) -> bool:
        """Takes a single step in the simulation. Part of the default schedule.

        Returns:
            bool: whether the simulation should continue
        """
        self._cycle += 1
        agents = self._ambient.get_agents()
        _LOGGER.debug("STEP(%s) - Agents(%s)", self._cycle, str(len(agents)))
        await self._step(agents)
        return self._ambient.is_alive

    async def _step_sync(self, agents: list[_Agent]) -> None:
        """Step all agents with sync points after `__sense__`, `__cycle__`, `__execute__`."""
        await _Future.gather([agent.__sense__(self._ambient) for agent in agents])
        await _Future.gather([agent.__cycle__() for agent in agents])
        await _Future.gather([agent.__execute__(self._ambient) for agent in agents])

    async def _step_async(self, agents: list[_Agent]) -> None:
        """Step all agents with a sync point at the end of each cycle."""
        futures = []
        futures.extend([agent.__sense__(self._ambient) for agent in agents])
        futures.extend([agent.__cycle__() for agent in agents])
        futures.extend([agent.__execute__(self._ambient) for agent in agents])
        await _Future.gather(futures)



"""Module implements logging functionality."""

import logging
from functools import wraps

# NOTE: _LOGGER is deprecated use LOGGER
__all__ = ("_LOGGER", "LOGGER", "Indent")

LOGGER = logging.getLogger(__package__)


class Indent:
    def __init__(self, spaces=2):
        self.logging_indent_level = 0
        self.logging_indent_spaces = " " * spaces

    def __call__(self, fun):
        @wraps(fun)
        def _log_with_indent(*args, **kwargs):
            with self:
                return fun(*args, **kwargs)

        return _log_with_indent

    def __enter__(self):
        self.logging_indent_level = self.logging_indent_level + 1
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logging_indent_level = self.get_current_indent_level() - 1

    def get_current_indent_level(self):
        return self.logging_indent_level

    def get_current_ident(self):
        return self.logging_indent_level * self.logging_indent_spaces


class IndentFormatter(logging.Formatter):
    def format(self, record):
        if not hasattr(record, "indent"):
            setattr(record, "indent", LOGGER.indent.get_current_ident())
        return super().format(record)


LOGGER.setLevel(logging.DEBUG)
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
formatter = IndentFormatter("%(levelname)s: %(indent)s%(message)s")
handler.setFormatter(formatter)
LOGGER.addHandler(handler)


def format_iterable(iterable, message, join="\n", indent=True):
    if indent:
        ind = "  " * LOGGER.indent.get_current_indent_level()
    else:
        ind = ""
    return join.join([f"{ind}{message}{x}" for x in iterable])


LOGGER.format_iterable = format_iterable
LOGGER.indent = Indent()
_LOGGER = LOGGER  # TODO _LOGGER is deprecated, it will be removed soon
